﻿:namespace Files ⍝ V1.12
⍝ File related commmands
⍝ 2015 11 12 Adam: moved help layout to framework
⍝ 2015 12 30 Adam: added Test
⍝ 2016 01 23 DanB: added help for find & replace
⍝ 2016 02 17 DanB: ]file.edit to work under nix (no use of ]caption)
⍝ 2016 03 18 DanB: modified help msg 

    ⎕ml←1 ⋄ ⎕io←1 ⋄ AllCmds←'Find' 'Replace' 'Edit' 'Touch' 'ToLarge'

    ∇ r←List;t
      r←{⎕NS ⍬}¨AllCmds
      r.Name←AllCmds
      r.Group←⊂'File'
      t←'L -folder= -options∊iI -types= -regex'
      r.Parse←('1',t)('2',t,' -backup[=]'),'1L -create' '1L' '1 -recursive -list -backup= -verbose'
     
      r[1].Desc←'Returns where a string is found in files'
      r[2].Desc←'Replace occurrences of strings by others in files'
      r[3].Desc←'Edit native file'
      r[4].Desc←'Make sure a native file exists'
      r[5].Desc←'Transform 32b component files into 64b'
    ∇

    ∇ r←level Help Cmd;with;by
      r←⊂List[AllCmds⍳⊂Cmd].Desc
      :Select Cmd
      :CaseList 'Find' 'Replace'
          :If 0=level
              r,←⊂'Arg: SearchString',(' ReplacementString'/⍨'R'=⊃Cmd),' [-folder=StartFolder] [-options=I] [-regex] [-types=exts]'
          :Else
              r,←⊂''
              r,←⊂'Argument',(' is text' 's are texts'⊃⍨1+'R'=⊃Cmd),' to be searched for in the relevant files'
              :If Cmd≡'Replace'
                  r,←⊂'-backup=    adds specified string (or by default ".bu") to the extension of the original files'
              :EndIf
              r,←⊂'-folder=    specifies a folder different than the current workdir'
              r,←⊂'-options=I  specifies case insensitivity'
              r,←⊂'-regex      interprets arguments as regular expressions'
              r,←⊂'-types=     specifies the file types (extensions) to be searched (default "dyalog,asmx,aspx,asax,apl")'
              r,←'' 'Examples:'
              by←5↓with←' with WXYZ'/⍨'R'=⊃Cmd
              r,←⊂'    ',Cmd,' ABC',with,' in .dyalog files under \temp'
              r,←⊂'        ]',Cmd,' ABC',by,' -folder=\temp'
              r,←⊂'    ',Cmd,' insensitive syntactic ABC',with,' in PHP files under \TMP'
              r,←⊂'        ]',Cmd,' \bABC\b',by,' -opt=i -folder=\tmp -typ=php -regex'
              r,←⊂'    ',Cmd,' ABC',with,' not within comments or text in the current workdir'
              r,←⊂'        ]',Cmd,' "(''[^'']*''|⍝.*$)?(?(-1)(*SKIP)(?!)|ABC)"',by,' -regex'
          :EndIf
     
      :Case 'Edit'
          :If 0=level
              r,←⊂'Arg: PathName [-create]'
          :Else
              r,←'' 'Argument is native file to be edited' '-create  creates the file first'
              r,←'' 'Examples:'
              r,←⊂'    Edit myfile.txt in the \TMP folder'
              r,←⊂'        ]file.',Cmd,' \tmp\myfile.txt'
              r,←⊂'    Create mynewfile.txt in the \TMP folder and edit it'
              r,←⊂'        ]file.',Cmd,' \tmp\mynewfile.txt -create'
          :EndIf
     
      :Case 'Touch'
          :If 0=level
              r,←⊂'Arg: PathName'
          :Else
              r,←'' 'Argument is native file to ensure existance of'
              r,←'' 'Example:' '    Check if abc.xyz exists under \TMP, and if not, create it'
              r,←⊂'        ]',Cmd,' \tmp\abc.xyz'
              r,←⊂'\tmp\abc.xyz created.'
          :EndIf
      :Case 'ToLarge'
          :If 0=level
              r,←⊂'Arg: StartFolder [-backup[=string]] [-list] [-recursive] [-verbose]'
          :Else
              r,←'' 'Argument is starting point where to transform files from.'
              r,←⊂'-backup=    adds specified string to the extension of the original files'
              r,←⊂'-list       only lists what would be done without actually doing it'
              r,←⊂'-recursive  applies the command recursively (includes all sub-folders)'
              r,←⊂'-verbose    provides additional details'
              r,←'' 'Example:'('    ]',Cmd,' /my/project -rec -ver -back=.D32')
              r,←⊂'"C:/my/project/t32.DCF" made into 64b format and backed up to "C:/my/project/t32.DCF.D32"'
              r,←'* "C:/my/project/talt/fo1.DCF" is already 64b' '...' '19 files modified'
          :EndIf
      :EndSelect
      r,←(level=0)/⊂']??FILE.',Cmd,' ⍝ for more info'
    ∇

    Open←{0::0⊣⎕←'*** Unable to open: ',⎕em ⎕en ⋄ 22::⍵ ⎕ncreate 0 ⋄ ⍵ ⎕ntie 0}

    ∇ r←Run(Cmd Args);ct;w;nt;n;tie;text;copy;ci;file;oldcaption;WIN
      r←0 0⍴''
      :If nt←3>ci←AllCmds⍳⊂Cmd
        ⍝ The utility functions are in this script
          ct←⎕SE.SALT.Load'tools\code\fileUtils -noname'
      :EndIf
     ⍝ We'll run one of them on these folders
      :If nt∧0≡w←⊂Args.Switch'folder'
          w←('∘'{1↓¨(v∊⍺)⊂v←⍺,⍵}⎕SE.SALT.Settings'workdir')⎕SE.SALTUtils.ClassFolder¨⊂''
      :EndIf
     
     ⍝ Is this NOT a regular expression?
      :If nt>Args.Switch'regex'
          Args.Arguments[1]←⊂'{}\.[^|]$(?*)+'{pat←⍵ ⋄ r←1+b←⍵∊⍺ ⋄ (pat b)←r∘/¨pat b ⋄ pat⊣((≠\b)/pat)←'\'}1⊃Args.Arguments
      :EndIf
     
      :Select ci
      :Case 1 ⍝ Find
          r←{'Total ',' found',⍨⍕+/{⍬⍴2⊃⎕VFI ⍵}¨⍵}w ct.showExpr¨⊂Args.(Arguments,types 1 options)
     
      :Case 2 ⍝ Replace
          r←{'Total ',' changes made',⍨⍕+/{⍬⍴2⊃⎕VFI ⍵}¨⍵}w ct.replExpr¨⊂Args.(Arguments,types 1 options backup)
     
      :Case 3 ⍝ Edit
          file←⎕SE.SALTUtils.fixFsep 1⊃Args.Arguments
          :If Args.create ⋄ →0/⍨0≡nt←Open file ⋄ ⎕NUNTIE nt ⋄ :EndIf
          :If 0≡text←{22::0 ⋄ ⎕SE.SALTUtils.GetUnicodeFile ⍵}file
              r←'** File not found'
          :Else
              (CR LF)←nt←⎕UCS 13 10 ⋄ r←'File not changed'
              copy←text←nt LF ⎕R(CR,⊂,CR)⍠'Mode' 'D'⊢text  ⍝ separate lines
              :If WIN←'Win'≡3↑1⊃'.'⎕WG'aplversion' ⋄ oldcaption←⎕SE.UCMD 1⌽'"caption editor "',1⊃Args.Arguments ⋄ :EndIf
              ⎕ED'text'
              :If WIN ⋄ {}⎕SE.UCMD'caption editor "',oldcaption,'"' ⋄ :EndIf
              :If copy≢text
                  :If ~⎕SE.SALTUtils.WIN ⋄ nt←LF ⋄ :EndIf
                  text←CR ⎕R nt⍠'Mode' 'D'⊢text
                  text ⎕SE.SALTUtils.PutUTF8File file ⋄ r←'File modified'
              :EndIf
          :EndIf
     
      :Case 4 ⍝ Touch
          r←1⊃Args.Arguments
          :Trap 0
              ⎕NUNTIE r ⎕NTIE 0
              r,←' exists.'
          :Case 22
              ⎕NUNTIE r ⎕NCREATE 0
              r,←' created.'
          :Else
              r←'*** Unable to access: ',⎕EM ⎕EN
          :EndTrap
     
      :Case 5 ⍝ ToLarge
          ⍝ Transform all 32b files in Args into 64b format
          ct←⎕SE.SALT.Load'tools\code\fileUtils -noname'
     
          ⍝ Get the list
          :If Args.recursive ⋄ r←1 ct.recFLib 1⊃Args.Arguments
          :Else
              r←⎕FLIB 1⊃Args.Arguments
              :If ⎕SE.SALTUtils.WIN
                  r←n⌽(r⌽⍨-n←+/∧\⌽' '=r),(r∧.≠'.')⍀1 4⍴'.DCF'
              :EndIf
          :EndIf
          ⎕←Args.verbose⌿r←↑Args.(backup list)∘ct.make64b¨{⍵↓⍨-⊥⍨' '=⍵}¨↓r
          r←(n←r∧.≠'*')⌿r
          →Args.list/0
          r←(⍕+/n),' files modified'
      :EndSelect
    ∇

    :SECTION Test
    ∇ r←Test dummy;T;cmd;n;tie;file;cf32b;Nix;grps;fol
      grps←List.Group
      r←⍬ ⋄ T←{⎕SE.UCMD(⊃grps[AllCmds⍳⊂cmd]),'.',cmd,' ',⍵}
      Nix←{⍺ ⎕NERASE ⍺ ⎕NTIE ⍵}
      fol←4↑file←'/tmp/filestst.dyalogtestfile'
      cf32b←¯86 13 13 2,(1\⍨¯4,5⍴1 ¯3),(31↑¯15 1 ¯3 1\¯112),¯3,n,(20↑5⍴4↑¯112),(68↑30⍴24↑¯12 ¯71 84 85 ¯59 3),111,(n←7↑3/¯1),16⍴8↑2
      ⎕NUNTIE ⎕NNUMS ⋄ ⎕FUNTIE ⎕FNUMS  ⍝ start clean
      :Trap 22 ⋄ file Nix 0 ⋄ :EndTrap ⍝ Erase if present
      ⎕SE.SALTUtils.makeDir'/tmp'
      ⎕NUNTIE tie⊣(∊'PleaSe' 'presS' 'Esc',¨⊂⎕UCS 13 10)⎕NAPPEND tie←file ⎕NCREATE 0
      :For cmd :In AllCmds
          :Trap n←0
              :Select cmd ⍝ Note: The order of the cases is important
              :CaseList 'Find' 'Replace'
                  n←(8↑T(8↑'"S" "s" P -opt=i'↓⍨8×'Find'≡cmd),' -fol=',fol,' -typ=',file↓⍨file⍳'.')≡'Total 2 '
              :Case 'Edit'
                  n←1 ⍝ (5↑T file)≡'File ' ⍝ Because it will need user action
              :Case 'Touch'
                  file Nix 0
                  n←(T file)≡file,' created.'
                  n∧←(T file)≡file,' exists.'
              :Case 'ToLarge'
                  ⍝ Replace text file with short-span component file:
                  ⎕NUNTIE tie⊣cf32b ⎕NAPPEND 83,⍨file ⎕NCREATE file Nix tie
                  n←(8↑T fol)≡'1 files '
                  n∧←64='S'⎕FPROPS tie←file ⎕FTIE 0
              :EndSelect
          :EndTrap
          r,←n
      :EndFor
      ⎕NUNTIE-|tie ⋄ ⎕FUNTIE|tie
      file Nix 0
     
    ∇
    :ENDSECTION
   
:endnamespace ⍝ Files  $Revision: 27267 $