﻿:namespace SaltInSpice ⍝ V1.10
⍝ Implement SALT commands within Spice
⍝ 2015 11 12 Adam: moved help layout to framework
⍝ 2015 12 30 Adam: added Test 
⍝ 2016 04 07 DanB: Settings example to use ∘

    ⎕ml← ⎕io←1
    Cnames←'Compare' 'RemoveVersions' 'List' 'Load' 'Save' 'Settings' 'Snap'  'Clean'

    ∇ r←List
      r←⎕NS¨(⍴Cnames)⍴⊂''
      r.(Group Parse)←⊂'SALT' ''
      r.Name←Cnames
      r[1].Desc←'Compare versions of 2 files'
      r[2].Desc←'Remove some or all versions'
      r[3].Desc←'List SALT files or folders in path'
      r[4].Desc←'Load an object (class or namespace or fn or var)'
      r[5].Desc←'Save an object (class or namespace or fn or var)'
      r[6].Desc←'Change local and global settings'
      r[7].Desc←'Save changes made to the workspace'
      r[8].Desc←'Remove all traces of SALT tags in the ws or in specific objects'
    ∇

    ∇ Û←Run(Ûcmd Ûargs);Ûriu
     ⍝ Simulate calling directly from the original ns
      Ûriu←{6::0 ⋄ ##.RIU}0
     ⍝ We can't use :With because if a space is returned
     ⍝ as result it will be lost in :endwith
      ⎕CS ##.THIS  ⍝ We know THIS has been set for us
      Û←⍎'⎕SE.SALT.',Ûcmd,' Ûargs' ⍝ Û vars for Snap
      :If Ûriu<Ûcmd≡'RemoveVersions'
          Û←0 0⍴''
      :EndIf
    ∇

    ∇ r←level Help Cmd
      (1↑Cmd)←⎕SE.Dyalog.Utils.ucase 1↑Cmd
      :If 'Clean'≡Cmd
          r←⊂'Removes all traces of SALT tags in the ws or in specific objects'
          :If 0=level
              r,←⊂'Args: [objects [-deletefiles]]'
          :Else
              r,←'' 'Argument is specific objects (default all) in the current namespace'
              r,←'' '-deletefiles also removes the files associated with the objects'' use'
              r,←'' 'Examples:' '    Remove all tags in all objects in the ws:' '       ]clean'
              r,←'    Remove tags in objects A and B and delete their files:' '       ]clean A B -deletefiles'
          :EndIf
      :Else
          r←{(+/∨\' '≠⌽⍵)↑¨↓⍵}⍎'⎕SE.SALT.',{(5×'SALT.'≡5↑⍵)↓⍵}Cmd,' ''?'''
          :If 0=level
              r←1 0 1/3↑r      ⍝ SALT returns too much help
              r[1]↓⍨←1+⍴Cmd    ⍝ Strip Cmd name from syntax info
              r[1],⍨←⊂'Args: ' ⍝ Add label to syntax info
          :Else  ⍝ More help (?? case)
              r↓⍨←1                   ⍝ Strip syntax info
              r←{⍵,⍨'    '/⍨'-'≡⊃⍵}¨r ⍝ Indent -modifier info
              r,←⊂'Examples:'
              :Select Cmd
              :Case 'Save'
                  r,←⊂'    Save "myobj" in <myfile> (not using a version number):'
                  r,←⊂'        ]save myobj /my/folder/myfile'
                  r,←⊂'    Start using version numbers with "myobj":'
                  r,←⊂'        ]save myobj -version'
                  r,←⊂'    Save "myobj" at a specific version number (here 12):'
                  r,←⊂'        ]save myobj -ver=12'
                  r,←⊂'    Save XYZ to folder ensuring folder is created if non existant:'
                  r,←⊂'        ]save XYZ /new/folder/xyz -makedir'
                  r,←⊂'    Save namespace NN and convert it to script form in the workspace:'
                  r,←⊂'        ]save NN  relative\path\nn  -convert'
              :Case 'RemoveVersions'
                  r,←⊂'    Remove version 12 of "myfile":'
                  r,←⊂'        ]removeversions /my/folder/myfile -version=12'
                  r,←⊂'    Remove versions above 12 of "myfile":'
                  r,←⊂'        ]removeversions /my/folder/myfile -version=>12'
                  r,←⊂'    Remove versions 12 to 21 of "myfile":'
                  r,←⊂'        ]removeversions /my/folder/myfile -version=12-21'
                  r,←⊂'    Remove versions starting at 12 of "myfile" except the last which will be renamed version 12:'
                  r,←⊂'        ]removeversions /my/folder/myfile -version=≥12 -collapse'
                  r,←⊂'    Remove all versions of "myfile" except the last one which will be renamed version 1:'
                  r,←⊂'        ]removeversions /my/folder/myfile -all -collapse'
                  r,←⊂'    Remove versionning for "myfile" and keep the last one:'
                  r,←⊂'        ]removeversions /my/folder/myfile -all'
              :Case 'Snap'
                  r,←⊂'    Save every new item in the ws to path specified; changes items are saved to their own location:'
                  r,←⊂'        ]snap C:\new\project'
                  r,←⊂'    Same thing but make sure the folder exists and do not prompt for confirmation of replacement:'
                  r,←⊂'        ]snap C:\new\project -makedir -noprompt'
                  r,←⊂'    Save only class 9 objects that start with "GUI" and convert namespaces to script form in the ws:'
                  r,←⊂'        ]snap /utils/gui -class=9 -pattern=GUI* -convert'
                  r,←⊂'    Save every new object in a file whose prefix is "UTIL" (eg fn <cat> goes to <UTILcat>):'
                  r,←⊂'        ]snap /basics -fileprefix=UTIL'
                  r,←⊂'    Save every item, including global variables which are stored in a separate file called <globals>'
                  r,←⊂'    and create a load fn named <lx> to recreate the ws:'
                  r,←⊂'        ]snap /my/wsx -vars=globals -loadfn=lx'
              :Case 'Load'
                  r,←⊂'    Load and define the object in a perticular file:'
                  r,←⊂'        ]load \tmp\myobject'
                  r,←⊂'    Load all version 3 of files starting with TM:'
                  r,←⊂'        ]load TM* -ver=3'
                  r,←⊂'    Bring script S, don''t define it, don''t keep a SALT trace and assign its reference to X:'
                  r,←⊂'        ]X←load \tmp\S -noname -nolink'
                  r,←⊂'    Disperse the objects in /tmp/ns in #.N:'
                  r,←⊂'        ]load /tmp/ns -disperse -target=#.N'
              :Case 'Compare'
                  r,←⊂'    Compare class in file C with object in workspace:'
                  r,←⊂'        ]compare /my/files/C -version=ws'
                  r,←⊂'    Compare versions 2 and 7 of file X; use +- as line symbols:'
                  r,←⊂'        ]compare \tmp\X -ver=2 7 -symb=+-'
                  r,←⊂'    Compare file A with file B, use always external pgm BC to compare:'
                  r,←⊂'        ]compare /tmp/A  /myfiles/B -use=/bin/BC -permanent'
              :Case 'Open'
                  r,←⊂'    Open disk explorer to show the contents of /tmp:'
                  r,←⊂'        ]open  /tmp'
                  r,←⊂'    Open file \my\file with Notepad (Windows):'
                  r,←⊂'        ]open  \my\file -with=notepad'
                  r,←⊂'    Open file X with program <vi>:'
                  r,←⊂'        ]open X -with=vi'
              :Case 'List'
                  r,←⊂'    List all files in \tmp and all files below:'
                  r,←⊂'        ]list \tmp -recursive'
                  r,←⊂'    List folders of folder Y:'
                  r,←⊂'        ]list \myfiles\Y -folders'
                  r,←⊂'    List versions and type of files in relative folder Z:'
                  r,←⊂'        ]list  Z  -type  -ver'
                  r,←⊂'    List the full (starting from #)  path of files:'
                  r,←⊂'        ]list  -full=2'
              :Case 'Settings'
                  r,←⊂'    Find the current settings for "workdir":'
                  r,←⊂'        ]settings workdir'
                  r,←⊂'    Overwrite the paths for User Commands:'
                  r,←⊂'        ]settings  cmddir  \my\ucmd1∘\my\ucmd2'
                  r,←⊂'    Set permanently the comparison program (for cmd Compare) to BestC:'
                  r,←⊂'        ]settings comp \programs\Bestc.exe  -permanent'
                  r,←⊂'    Remove the requirement to answer prompts when saving objects:'
                  r,←⊂'        ]settings edprompt 0'
              :EndSelect
          :EndIf
      :EndIf
      :If 0=level ⋄ r,←⊂']??SALT.',Cmd,' ⍝ for more info' ⋄ :EndIf
    ∇

    :Section TEST

    ∇ r←Test dummy;T;file;testvar;Eq;ext;old
      r←⍬ ⋄ T←{0::0 ⋄ ⎕SE.UCMD'SALT.',⍺,' ',⍵}
      Eq←{⊃≡/~∘'/\'¨⍺ ⍵~¨⊂'/\'}
      testvar←⎕D
      old←'Settings'T'varfmt xml'
      r,←'/tmp/testvar.dyalog'Eq'Save'T'testvar /tmp/ -nop'
      ⎕EX'testvar'
      r,←(testvar≡⎕D)∧'testvar'≡'Load'T'/tmp/testvar'
      r,←'testvar'≡2⊃2⌷'List'T'/tmp/testvar'
      r,←2=+/'testvar'⍷'Compare'T'testvar -ver=ws'
      r,←(≢'List'T'/tmp/testvar')='RemoveVersions'T'/tmp/testvar -all -nop'
      r,←'xml'≡'Settings'T'varfmt ',old
      r,←1⊣'Snap'T''⍝ not sure how it works
      r,←∨/'Cleanup done'⍷'Clean'T'' ⍝ delete what snap made
     
    ∇
    :EndSection

:Endnamespace ⍝ SaltInSpice  $Revision: 27420 $
