﻿:namespace  Spice ⍝ V1.23
⍝ UCMD commands Script
⍝ 2015 09 25 DanB: added GetVersion
⍝ 2015 11 02 DanB: added more info to ]version and -extended
⍝ 2015 12 30 Adam: added 8.1 and 10 handling to ]version
⍝ 2016 02 18 Adam: documented ]version -extended, various Help fixes, moved ]version to TOOLS

    Cmds←'ULoad' 'UReset' 'UDebug' 'Version'  'UMonitor' 'Refresh'  ⍝ 'UHelp'?

    ⎕ML←1 ⋄ ⎕IO←1 ⋄ NL←⎕av[4 3] ⍝ cannot use ⎕ucs 13 10 on Classic Unix

    rtb←{⍺←' ' ⋄ (-⊥⍨⍵∊⍺)↓⍵} ⋄ DFLAGS←'d1c48KSWwh'

    ∇ r←List
      r←{⎕NS ⍬}¨Cmds
      r.Name←Cmds
      r.(Group Parse)←⊂'UCMD' ''
      r[1].(Desc Parse)←'Load a User command''s code into the workspace' '1'
      r[2].Desc←'Reload all user command definitions in the workspace' ⍝ this cmd is dealt with in SALTUtils
      r[3].(Desc Parse)←'Enable debugging of User Commands'('1s -permanent -flags[=]')
      r[4].(Desc Parse Group)←'Report various version numbers (e.g. APL/SALT/UCMD)' '1SL -extended' 'TOOLS'
      r[5].(Desc Parse)←'Monitor code execution' '1s -var= -report[=]'
     ⍝ This command should be in the SALT class:
      r[6].(Desc Parse Group)←'Refresh objects in the ws with the most recent version' '-noprompt' 'SALT'
    ∇

    ∇ r←Run(Cmd Args);z;la;t;fn;unav;state;ref;cl;setf;oo;un;rel;v;a;noa;fl;nof;obj;file;data;p6;ver;type;os;ext
      r←0 0⍴''
      :Select Cmd
     
      :Case 'ULoad'
          r←'No such command found' ⋄ v←1↑⍴t←⎕SE.Dyalog.SALT.List ⋄ un←{⍺,'.',⍵}/t[;2,⍨⍳'.'∊ref←(1⊃Args.Arguments)~' ']
          :If p6←⎕SE.SALTUtils.UVersion≤1.36 ⋄ un←⎕SE.Dyalog.Utils.lcase¨t[;1] ⋄ :EndIf
          :If ∨/z←un∊⊂cl←⎕SE.Dyalog.Utils.lcase ref←(1⊃Args.Arguments)~' '
          :OrIf 1=+/z←cl∘≡¨(⍴cl)↑¨un
              cl←⎕IO+9.1=⎕NC⊂r←⍕⎕SE.SALT.Load'"',((7-p6)⊃fn←t[z⍳1;]),'" -target=',⍕##.THIS
              r←'The source code for command "',((2-p6)⊃fn),'" has been loaded in ',(cl⊃'class' 'namespace'),' ''',r,''''
          :EndIf
     
      :Case 'UDebug'
          r←'Is ',state←(0 1⍳⎕SE.SALTUtils.DEBUG)⊃'OFF' 'ON' 'ON+'
          →0/⍨(noa←' '∧.=a←⍕Args.Arguments)∧nof←0≡fl←Args.flags
          '-permanent does not apply to flags'⎕SIGNAL 11/⍨Args.permanent>nof
          r←'Unable to set (enter ON or OFF)'
          :If ∨/t←'off' 'on' ''∊⊂⎕SE.Dyalog.Utils.lcase⌽∘rtb⍣2,a~' '
              :If nof
                  r←'Was ',state
                  ⎕SE.SALTUtils.DEBUG←t[2]
                  :If Args.permanent ⋄ {}⎕SE.SALT.Settings'debug ',(⍕t[2]),' -permanent' ⋄ :EndIf
              :Else
         ⍝ Set the interpreter (DEBUG) flags (OFF means remove them, ON adds them, neither queries)
     
⍝ The flag are
⍝ h additional checks
⍝ w wscheck each desk calc line
⍝ W wscheck each line of apl
⍝ S shuffle 'n wscheck each getspace
⍝ M shuffle and wscheck each trip to message q
⍝ K log all input keys to apllog
⍝ 4 ?
⍝ c wscheck after each callback
⍝ 1 or d: debug
     
                  ((fl='d')/fl)←'1'  ⍝ d ≡ 1 flag
                  r←(1+¯1↑t)⊃'Was: ' 'Is: '
                  setf←{2 ⎕NQ'.' 'SetDFlags'⍵}
                  r←r,t[2]{n←2⍴⍨⍴tags←⌽1↓DFLAGS
                      old←bits←n⊤setf 0  ⍝ find current state
                      ((tags∊⍵)/bits)←⍺
                      x←setf 2⊥bits      ⍝ reset to new value
                      1⌽'""',old/tags}fl,(fl≡1)/DFLAGS
              :EndIf
          :EndIf
     
      :Case 'Version'
          r←Version Args
     
      :Case 'UReset'
          ⍝ Execution of UReset is handled in the framework; it appears here for the HELP only
          ⍝ and in case a shorter form of 'ureset' has been entered:
          ResetUCMDcache 1
          t←⍕1↑⍴⎕SE.Dyalog.SALT.List ⋄ r←t,' commands reloaded'
     
      :Case 'UMonitor'
          state←(0 1⍳⎕SE.SALTUtils.MONITOR)⊃'OFF' 'ON' 'ON+'
          :If 0≢file←Args.report
          ⍝ Produce a report of code coverage
              :If ⎕SE.SALTUtils.MONITOR=0
              :OrIf 0=⎕NC v←⎕SE.SALTUtils.MONITORNAME
              :OrIf 0∊⍴⍎v
                  'No data accumulated'⎕SIGNAL 6
              :EndIf
              0 ⎕THIS ⎕SE.SALT.Load'tools/code/codecov' ⍝ bring in the code
              data←,0 1↓¯2 CodeCov.CCinit ⎕IO⊃¨⍎v
              r←data CodeCov.CC 2⊃¨⍎v
              :If file≢1
                  r ⎕SE.SALTUtils.PutUTF8File file
                  r←(⍕r),' bytes written to ',file
              :EndIf
          :ElseIf 0≡v←Args.var ⋄ r←'Is ',state
          :Else ⋄ rel←~'#⎕'∊⍨1↑v
              'MONITOR cannot be used from an unnamed NS'⎕SIGNAL 11/⍨rel∧'['∊t←⍕##.THIS
              v←(rel/t,'.'),v
              'invalid variable name'⎕SIGNAL 11↓⍨(⎕NC v)∊0 2
              ⎕SE.SALTUtils.MONITORNAME←v
              r←'''',v,''' will be used for monitoring'
              ⎕SE.SALTUtils.MONITOR←1
          :EndIf
          →0↓⍨⍴Args.Arguments
          :If ∨/oo←'off' 'on'∊⎕SE.Dyalog.Utils.lcase¨Args.Arguments
              r←'Was ',state
              :If ⎕SE.SALTUtils.MONITOR←oo[2] ⍝ turn ON? clear data
              :AndIf 2=⎕NC t←⎕SE.SALTUtils.MONITORNAME
                  ⍎t,'←⍬'
              :EndIf
          :Else ⋄ r←'Unable to set (enter ON or OFF)'
          :EndIf
     
      :Case 'Refresh'
          r←(~Args.noprompt)refreshWS #
     
      :EndSelect
    ∇

    ∇ r←Version Args;ext;fn;ok;os;t;type;unav;v;ver;z;⎕USING
    ⍝ Return version information about a number of items
     
      :If 0<⍴fn←{(+/∧\' '=⍵)↓⍵}1⊃1↑Args.Arguments ⍝ are we trying to find the version of a ws?
          :If '⎕wsid'≡⎕SE.Dyalog.Utils.lcase fn ⋄ fn←{⍵,(~'.'∊⍵)/'.DWS'}⎕WSID ⋄ :EndIf
          :Trap 22
              z←⎕SE.SALTUtils.WIN>'.'∊fn  ⍝ under Win add extension if missing
              t←('"'~⍨fn,z/'.DWS')⎕NTIE 0 ⋄ r←1⍕0.1⊥⌽⎕NREAD t,83 2 4 ⋄ ⎕NUNTIE t
          :Else
              ('workspace ',fn,' is not found')⎕SIGNAL 22
          :EndTrap
      :Else ⍝ show various versions
        ⍝ Dyalog's
          type←(32×1+∨/'-64'⍷v←1⊃ver←'.'⎕WG'aplversion'){' ',(⍕⍺),'-bit ',⍵⊃'Unicode' 'Classic'}1+82=⎕DR''
          r←⊂'Dyalog'(({(3>+\'.'=⍵)/⍵}2⊃ver),type,', BuildID ',{0::'?' ⋄ 2 ⎕NQ'.'⍵}'getbuildid')
        ⍝ The OS
          :If 0∊⍴os←('W'=1↑v)/v ⍝ non Win?
              os←{0::'' ⋄ ∊⎕SH ⍵}'uname -srv','mM'[1+'AIX'≡3↑v]
          :Else                 ⍝ Windows
              ⎕USING←'System'
              :Trap 6
                  t←{(2>+\'.'=⍵)/⍵}⍕Environment.OSVersion
                  os,←', ',(~∨\'Microsoft Windows NT'⍷t)/t ⍝ remove NT reference
              :EndTrap
              os←GetVersion
          :EndIf
          r,←⊂'OS'os
        ⍝ SALT & UCMDs
          t←unav←'(unavailable)'
          :If 9=⎕SE.⎕NC'SALT' ⋄ t←⍕⎕SE.SALT.Version ⋄ :EndIf
          r,←⊂'SALT't
          r,←⊂'UCMD'(⍕⎕SE.SALTUtils.UVersion)
        ⍝ .Net: there are the Version in use, the bridge and dyalognet
          v←unav
          :Trap 6
              :If Args.extended⊣t←⍬
                  t←((7↑¨t)∊'bridge1' 'dyalogn')/t←⍕¨AppDomain.CurrentDomain.GetAssemblies
                  t←{(~∨\', Culture'⍷⍵)/⍵}¨t
              :EndIf
              v←↑t,⍨⊂{(2>+\'.'=⍵)/⍵}⍕Environment.Version
          :EndTrap
          r,←⊂'.NET'v
     
          :If Args.extended
              ext←'Conga'unav
              :Trap 0
                  'DRC'⎕CY'conga'
                  :If ok←0=1↑v←2 DRC.Init 0 ⍝ try calling with a 2 (load copy) left arg
                      t←⊂6↓2⊃v   ⍝ split on various sections, e.g. GnuTLS
                      v←v⊂⍨↑∨/'Dynam' 'Copyright' ' built' 'GnuTLS' 'Built'⍷¨⊂v←2 2⊃DRC.Describe'.'
                      v←,/1 0 1/2 3⍴v ⍝ remove copyright sections
                      t,←v
                  :EndIf
                  :If ok ⋄ ext[2]←⊂↑t ⋄ :EndIf
              :EndTrap
              r,←⊂ext
     
              ext←'SQAPL'unav
              :Trap 0
                  'SQA'⎕CY'sqapl'
                  :If ok←0=1↑v←SQA.Init'' ⋄ t←⊂(~∨\'Using'⍷t)/t←6↓2⊃v
                      v←v⊂⍨↑∨/'Version' 'Copyright' ' built' 'IniFile'⍷¨⊂v←2⊃SQA.Describe'.'
                      v←,/1 0 1/2 3⍴v,'' '' ⍝ remove copyright sections
                      t←t,¯2 0↓¨v
                  :EndIf
                  :If ok ⋄ ext[2]←⊂⎕FMT↑t ⋄ :EndIf
              :EndTrap
              r,←⊂ext
          :EndIf
      :EndIf
         ⍝ Other info such as computer name, OS, 64bOS, #CPUs could be added
      r←⍕↑r
    ∇

    isYes←{~⍺:1 ⋄ ⎕←⍵ ⋄ n←⍴⍞←'? [Yes/No/Stop]: ' ⋄ ∨/'sS'∊r←1↑n↓⍞:⎕signal 901 ⋄ ∨/'yY'∊r}

    remVer←{1≥+/b←(∨\r∊'/\')<'.'=r←⌽⍵:⍵ ⋄ (,1)≢1⊃⎕vfi r/⍨~c←1≠+\b:⍵ ⋄ ⌽c/r}

    ∇ {n}←confirm refreshWS tgt;lst;sons;file;srcws;t;b;obj;cload;call1;⎕TRAP;⎕PW
    ⍝ This fn is used to update all objects starting at tgt
      ⎕PW←999 ⋄ n←0 ⍝ number of changes made
      :If call1←0=⎕NC'LIST' ⋄ ⎕SHADOW'LIST' ⋄ LIST←⍬ ⋄ ⎕TRAP←901 'C' '→End' ⋄ :EndIf ⍝ avoid recursion, keep track of nss done
      LIST,←tgt
      cload←{(o f p)←⍵ ⋄ 922 22::0⊣⎕←'*** File ',f,' NOT FOUND for object ',o ⋄ ⍺≢⎕SE.SALT.Load f,p}
     
      :If 0≢srcws←{0::0 ⋄ ⎕SRC ⍵}tgt ⍝ chk scripted nss
          :If 9=tgt.⎕NC'SALT_Data'
              :If srcws cload tgt(file←tgt.SALT_Data.SourceFile)' -source'
              :AndIf confirm isYes⍕'** Confirm overwrite ',tgt,' with contents of file ',remVer file
                  n+←⍴,⎕←0 tgt.## ⎕SE.SALT.Load file ⍝ we need to redo the load to tag the object
              :EndIf
          :EndIf
      :Else ⍝ must be a regular ns
     
          :If 2=#.⎕NC'SALT_Var_Data.VD'
          :AndIf 1∊b←#.SALT_Var_Data.VD[;⎕IO]{⍵∘≡¨{⍵↓⍨-1+⊥⍨'.'≠⍵}¨⍺}{df←⍵.⎕DF ⎕NULL ⋄ (⍵.⎕DF df)⊢⍕⍵}tgt
              :For obj file :In ↓b⌿#.SALT_Var_Data.VD[;⍳2]
                  :If (⍎obj)cload obj file' -noname -nolink'
                  :AndIf confirm isYes⍕'** Confirm overwrite ',obj,' with contents of file ',remVer file
                      n+←⍴,⎕←0 tgt ⎕SE.SALT.Load file ⍝ we need to redo the load to tag the object
                  :EndIf
              :EndFor
          :EndIf
     
          :For obj :In lst←tgt.⎕NL-3 4
              :If 2≡≡srcws←tgt.⎕NR obj
              :AndIf ~0∊⍴t←⎕SE.SALTUtils.fnData srcws obj
              :AndIf ({(3≠⍴⍵)∨' }'≢1⊃⌽⍵:⍵ ⋄ ⍵[1],¨1↓¨⍵[3]}⎕SE.SALTUtils.remTag srcws)cload obj(file←t.SourceFile)' -source'
              :AndIf confirm isYes⍕'** Confirm overwrite ',obj,' with contents of file ',remVer file
                  n+←×⍴,⎕←0 tgt ⎕SE.SALT.Load file
              :EndIf
          :EndFor
     
          :If 0<⍴lst←tgt.⎕NL-9
          :AndIf 0<⍴lst←LIST~⍨tgt⍎¨lst
            ⍝ Recurse thru non-scripted children
              :If 0<⍴sons←{⍵/⍨~1∊¨(⍕¨⍵)∊¨⊂'[('}lst
                  n+←+/∊confirm refreshWS¨sons
              :EndIf
          :EndIf ⍝ all nss
      :EndIf
     End:
      :If call1 ⋄ n←(⍕n),' objects refreshed' ⋄ :EndIf
    ∇

    ∇ r←level Help Cmd
      :Select Cmd
      :Case 'ULoad'
          r←,⊂'Load in the workspace the namespace associated with a User cmd.'
          r,←⊂'The argument is the command name found in the namespace.'
          →level↓end
          r,←'' 'Example, load the script where the CD cmd resides:'
          r,←('      ]',Cmd,' cd')'The source code for command "CD" has been loaded in namespace <#.Summary>'
      :Case 'UReset'
          r←⊂'Forces a reload of all user commands, to ensure that parsing information is up-to-date (after changing parsing rules for a command).'
     
      :Case 'UDebug'
          r←''(']',Cmd,' on       Enables debugging of user commands')
          r,←⊂']',Cmd,' off      Causes all errors in user commands to be trapped (this is the default)'
          r,←⊂' -permanent  makes the setting permanent across APL sessions'
     
          r,←'' 'NB: With debugging enabled, a trailing "-" on a user command allows easy tracing of the call'
          r,←'' ' -flags  facilitates debugging within the user command processor and should only be used as directed by Dyalog.'
     
      :Case 'Version'
          r←⊂'Reports various version numbers (APL, SALT, UCMD, .NET)'
          r,←⊂'If given the path to a workspace as argument it will report the minimum version necessary to )LOAD it.'
          r,←⊂' -extended  gives additional version information for .NET, Conga, and SQAPL.'
     
      :Case 'UMonitor'
          r←⊂'Turn monitoring ON or OFF (the argument)'
          r,←⊂'Monitoring is done with ⎕MONITOR for all the user commands.'
          r,←⊂'⎕MONITOR values are kept in #.UCMDMonitor after each invocation of a command.'
          r,←⊂'The data can be later used for producing code coverage reports for example.'
          r,←⊂'-var= sets the name of the variable to store the result'
          r,←⊂'If var is specified monitoring is turned on automatically'
          r,←⊂'-report creates a report to the screen, -report=file sends it to file'
          r,←⊂'Turning monitoring ON discards any previous data'
          r,←'To see examples type'('      ]??',Cmd)
          :If ×level
              r,←'Examples:'('      ]',Cmd,' ON')'      ⍝ call user command XXX' '      ]XXX'
              r,←⊂'     ]',Cmd,' -report'
          :EndIf
     
      :Case 'Refresh'
          r←⊂'Refreshes all SALTed objects by reloading the lastest version.'
          r,←⊂'-noprompt  skips the confirmation'
      :EndSelect
     end:r←↑r
    ∇

    ∇ R←GetVersion;gv;V;⎕IO;⎕ML;minor;nul
      ⎕IO←⎕ML←1
      'gv'⎕NA'U4 kernel32∣GetVersion'
     ⍝ OS is specified by high-order (first) bit and low-order byte
      :Select (V←(32⍴2)⊤gv)[1 30 31 32]
      :Case 0 0 1 1 ⋄ R←'Windows NT 3.51'
      :Case 0 1 0 0 ⋄ R←'Windows NT 4.0'
      :Case 0 1 0 1 ⋄ R←'Windows 2000 or Windows XP'
      :Case 1 1 0 0 ⋄ R←'Windows 95, Windows 98, or Windows Me'
      :Case 0 1 1 0
    ⍝ Major version of 6. Look at minor version.
          minor←2⊥V[16+⍳8]
          R←'Vista or Windows Server 2008' '7 or Windows Server 2008 R2' '8 or Windows Server 2012'
    ⍝ Note that 8, 8.1, and 10 (and probably all future versions) report 8.0⌈whatever the application manifest calls for
          R,←'8.1 or Windows Server 2012' '10 beta' ⍝ pre-releases of 10 reported 6.4
          R←'Windows ',(R⊃⍨0 1⍳minor),(' or newer'/⍨2≤minor)
      :Case 0 0 0 1 ⋄ R←'Windows 10 alpha' ⍝ earliest builds reported 9.0
      :Case 0 0 1 0 ⋄ R←'Windows 10 or Windows Server 2016' ⍝ only if application manifest says 10
      :EndSelect
      :Trap 0
          'gv'⎕NA'u ntdll.dll|RtlGetVersion ={ U U U U U T[128]}'
          V←2 6⊃V←gv⊂276 0 0 0 0(128⍴' ')
         ⍝ Some versions return Wide chars
          :If 1<+/V=nul←⎕UCS 0
              V←((⍴V)⍴1 0)/V
          :EndIf
          R,←', ',(¯1+V⍳nul)↑V
      :EndTrap
    ∇

:Endnamespace ⍝ Spice  $Revision: 26667 $