﻿:namespace  inout ⍝ V1.14
⍝ Transfer objects in and out of the ws
⍝ 2015 05 21 Adam: NS header
⍝ 2016 07 21 DanB: changed Help

    (⎕IO ⎕ML)←1


    AllCmds←'in' 'out'

    APLs←'APL2PC' 'APL2MF' 'VSAPL' 'APLX' 'DYALOG' 'APLPLUS' 'STSCMF' 'SHARPMF' 'SHARPX'
    Surr←'A2K'    'AP2'    'AP2'   'APX'  'DYW'    'A2K'     'A2K'    'SAM'     'SAX'
    Xext←'xiw'    'xiw'    'xiw'   'xpw'  'xdw'    'xsw'     'xsw'    'xmw'     'xuw'
    Xext←↑Xext((¯1↓¨Xext),¨'f') ⍝ component files

    ∇ r←List
      r←⎕NS¨(⍴AllCmds)⍴⊂''
      r.Name←AllCmds
      r[1].Desc←'Recreate a ws from a text file describing it' ⍝ descriptions
      r[2].Desc←'Copy the current ws into a text file'
      r.Group←⊂'Transfer'                                 ⍝ same group
     ⍝ Parsing rules for each:
      r[1].Parse←'1 -atf -fliplu -file[=] -list -noam -obj= -q -range= -replace -trans[=]0 1 2 -apl=',⍕APLs
      r[2].Parse←'1 -atf -obj= -file[=] -range= -lock= -q'
    ∇

    ∇ r←Run(Cmd A);arg;nars;files;q;Cn;ext;msg;folder;name;n;findExt;lc;x;b;apl
    ⍝ Transfer code to/from files
      Cn←AllCmds⍳⊂Cmd ⋄ lc←⎕SE.SALTUtils.(lCase⍣WIN) ⋄ findExt←{⍵↑⍨-⊥⍨b∧0∊b←'.'≠⍵}
      (folder name)←⎕SE.SALTUtils.{'/\'splitLast ⍵}'"'~⍨arg←1⊃A.Arguments
      :If (1=Cn)∧0∊⍴ext←findExt name
      :AndIf 0∊⍴ext←(⊂(1+A.file),APLs⍳⊂A.apl)⊃Xext,⊂''
         ⍝ In input mode, if the extension has been omitted, we try to guess
          files←'f'⎕SE.SALTUtils.Dir(arg~'"'),'.*'   ⍝ all files in that folder
          ext←lc∘findExt¨files                       ⍝ their extensions
          x←ext∩(1+A.atf)⊃(,Xext⌿⍨⍲\2⍴0≡A.file)(⊂'atf')
          :If 1=n←⍴x                                 ⍝ only one possibility?
         ⍝ If there is only one file it is easy
              (ext arg)←(ext⍳x)⊃¨ext files ⋄ arg←folder,'/',arg
          :Else
              msg←(1+×n)⊃'file not found'('ambiguous filename (',(⍕n),' possibilities)')
              msg ⎕SIGNAL 22
          :EndIf
      :EndIf
     
      :If (ext≢'NARS')∧∨/b←'''" -'∊arg
         ⍝ The fns below will reparse the line so we need to surround the arg properly with quotes
          arg←q,((1+q=arg)/arg),q←''''
      :End
     
      :Select Cn
      :Case 1 ⍝ IN
         ⍝ We accept ATF, NARS and extended format files (DYW, APX, etc.)
          :If A.atf∨'atf'≡ext←ext,(A.atf∧0∊⍴ext)/'atf'
              ⎕FIX ⎕SE.SALT.Load'lib\atfin -source'
              ##.THIS ⍙⍙.ATFIN arg,('.'∊arg)↓'.',A.Propagate'fliplu apl' ⍝ ensure '.' in name
          :ElseIf ext≡'nars'
              ##.⎕CY'xfrcode.dws'
              ##.THIS ##.xfr.nars.makeWS arg
          :Else
              :If APLs∊⍨⊂apl←A.apl~0 ⋄ apl←' -apl=',(b⍳1)⊃Surr ⋄ :EndIf
              ##.⎕CY'xfrcode.dws'
              ##.THIS ##.xfr.∆xfrfrom arg,apl,A.Propagate'file list noam obj q range replace trans'
          :EndIf
     
      :Case 2 ⍝ OUT: we export to ATF if the extension is ATF or -atf has been supplied
          :If A.atf∨(⊂ext)∊'atf' 'ATF'
          ⍝ Before proceeding we check if there are any exotic objects in the ws.
          ⍝ We should also check for unusual fns accepting or returning multiple args but we don't now.
              :If 0<n←+/~2.1 3.1∊⍨##.THIS.⎕NC ##.THIS{0≡⍵:⍺.⎕NL-⍳9 ⋄ ⎕ML←3 ⋄ (','≠⍵)⊂⍵}A.obj
                  ⎕←'*** There are ',(⍕n),' objects impossible to transfer in ATF format.'
                  ⎕←'    Consider using the non-ATF format instead.'
              :Else
                  ⎕FIX ⎕SE.SALT.Load'lib\atfin -source'
                  ##.THIS ⍙⍙.ATFOUT arg,A.Propagate'obj'
              :EndIf
          :Else ⍝ must be extended form
              ##.⎕CY'xfrcode.dws'
              ##.THIS ##.xfr.∆xfrto arg,A.Propagate'file range lock obj q'
          :EndIf
      :EndSelect
      r←0 0⍴''
    ∇

    ∇ r←lev Help Cmd;o;i;ix;ox;CR;table
      :If Cmd≡'in'
          :Select lev
          :CaseList 0 1
              i←⊂'Bring objects defined in a transfer file.'
              ix←'The source APL can be specified through the -apl switch or the file extension can tell the program how to handle the input file.'
              ix,←' An extension of .atf will use the ATF format (APL Transfer Format).'
              i,←⊂ix,' With ATF the acceptable APLs (-apl values) are:'
              i,←⊂⍕¯1↓APLs
              ix←'With them, if the source APL was using underscored letters you can switch them with lowercase letters by using the -fliplu switch.'
              i,←ix'' 'Other formats are possible. Nars2000 is one of them.'
              i,←⊂'The APL Extended Format is another format. Several switches allow to fine tune the process. The file extensions associated with each APL are:'
              i,←↓⍕APLs⍪Xext
              i,←⊂'The x?w extensions relate to workspaces, the x?f extensions relate to files.'
              i,←⊂'See ]??in for details'
              r←↑i ⋄ →lev↓0
     
              i←¯1↓i ⍝ remove last ]??in
              ix←'ATF files are files used historically to transfer APL code. '
              ix,←'They were introduced in the 80s and are sometimes used to recover code from other '
              ix,←'APL vendors. They don''t always follow the same format and can be difficult to deal with.'
              i,←''ix'' 'Example: bring in an ATF file generated by another vendor' '    ]in \qa\xfr\aplx\ttt3d -atf'
              i,←'' 'Transfer files can also be generated in those APLs in a different serialized format.'
              i,←'' 'Example: retrieve all objects from an APL/PC transfer file and translate code:'
              i,←⊂'    ]in \qa\xfr\aplpc\starmap.xsw -tr'
              i,←'' 'Retrieve a ws from a Sharp APL (Unix) transfer file and define translate code:'
              i,←⊂'    ]in \qa\xfr\Sharp\utils.xuw -tr=2'
              i,←'' 'Retrieve some character and numeric data from a Dyalog transfer file:'
              i,←⊂'    ]in \qa\xfr\dyw\misc.xdw -obj=CN'
              i,←'' 'Retrieve all objects in a NARS workspace:' '    ]in \qa\xfr\nars\x1.ws.nars'
              i,←⊂'Note that you cannot retrieve individual objects from a NARS ws nor perform translation.'
              i,←'' 'For a complete description of this command do' '      ]???in'
              r←↑i
     
          :Case 2
              'xfr.∆describe'⎕CY'xfrcode.dws'
              r←'to/from'⎕R'from'⊢∆describe
              r,←CR,'For an example of transformations for APLX do',(CR←⎕UCS 13),'      ]????in'
     
          :Else
              'xfr.∆equivalence'⎕CY'xfrcode.dws' ⋄ CR←⎕UCS 13
              ix←1⍳⍨':APX'⍷∆equivalence ⍝ find APLX section
              r←ix↓∆equivalence
              r←(1+1⍳⍨CR CR⍷r)↓r ⍝ remove header
              r←(¯1+1⍳⍨CR CR':'⍷r)↑r ⍝ drop next sections
              r←'%'⎕R' → '⊢'"\d+"'⎕R{92 3::'?' ⋄ ⎕UCS 9031 9032 9047 9040[11 12 9 7⍳2⊃⎕VFI ⍵.Match~'"']}r
              r←↑1↓¨(r=CR)⊂r ⍝ turn into a matrix
              i←↓3 ⎕SE.Dyalog.Utils.showCol r
              r←'Here is the list of all the changes made for APLX:' '',i
              r,←'' 'For a complete description of this command do' '      ]???in'
     
          :EndSelect
     
      :ElseIf Cmd≡'out'
          :If lev=0
              o←⊂'Write objects to an ATF or serialized file.'
              o,←⊂'If the format is ATF only the variables and programs in the current namespace are exported.'
              o,←⊂'For the serialized format everything but forms (instances) and program references is exported.'
              o,←⊂'You can specify which objects to be written out with the -obj modifier'
              r←↑o,⊂'See ]??out for details.'
     
          :ElseIf lev=1
              o←⊂'Write objects to an ATF or serialized file.'
              o←,'' 'Example:' 'Write current workspace contents to file \tmp\aplp in ATF format:'
              o,←'      ]out \tmp\aplp  -atf' ''
              o,←'Serialize 3 objects (not ATF):' '      ]out \tmp\myws -obj=A Bc def'
              o,←'' 'To retrieve the objects in another APL, )LOAD the xfrpc ws on that APL and do' '      ∆xfrto ''\tmp\myws''   ⍝ extension .xdw will be added'
              o,←'' 'To retrieve the objects in (possibly another version of) Dyalog APL do' '      ]in  \tmp\myws.xdw   ⍝ for example'
              o,←'' 'For a complete description of the code do' '      ]???out'
              r←↑o
     
          :Else
              'xfr.∆describe'⎕CY'xfrcode.dws'
              r←'to/from'⎕R'to'⊢∆describe   ⍝ M11146
          :EndIf
      :EndIf
    ∇

:Endnamespace ⍝ inout  $Revision: 27923 $