﻿:Namespace caption ⍝ V2.03
⍝ Query or change window captions
⍝ 2015 10 27 Adam: Changed class name
⍝ 2015 11 12 Adam: moved help layout to framework
⍝ 2016 03 17 DanB: fixed case when not found
⍝ 2016 05 05 DanB: added SysTray for V15


    :SECTION FUNCTIONS

    ∇ r←List
      →0↓⍨⍴r←⎕NS¨⎕SE.SALTUtils.WIN⍴⊂⍬
      r.Name←⊂'Caption'
      r.Group←⊂'MSWin'
      r.Desc←⊂DESC
      r.Parse←⊂'2SL'
    ∇

    ∇ r←level Help Cmd;l;kw
      r←⊂DESC
      :If 0=level
          r,←⊂'Args: window ["new caption"]'
          r,←⊂']??',Cmd,' ⍝ for more info'
      :Else
          r,←'' '1st argument is one of the following window names:'
          :If level∊1 3
              r,←⊂'    ',windowString
              :If 3=level
                  r,←'' 'Examples:'
                  r,←⊢⊢⊢'    Change the session title to "ABC":'('        ]',Cmd,' session ABC')
                  r,←⊢⊢⊢'    Reset the session title:'('        ]',Cmd,' session ""')
              :EndIf
          :Else ⍝ 2=level
              l←⊢⊂'main Dyalog APL session window'
              l,←⊂'Editor window'
              l,←⊂'Message Box when an object cannot be fixed'
              l,←⊂'The Workspace Explorer tool'
              l,←⊂'Dialog box when one or more objects cannot be re-instantiated upon loading a workspace'
              l,←⊂'Status window'
              l,←⊂'Refactor as Method/Field/Property window displayed by the Editor'
              l,←⊂'Event Viewer'
              l,←⊂'Find/Replace dialog box'
              l,←⊂'Exit dialog box that is displayed when the user closes the main session'
              l,←⊂'Find Objects tool'
              l,←V15/⊂'System Tray'
              r,←⊢'   '∘,¨↓⍕windowValues,⍪l
          :EndIf ⍝ Continue 1≤level
          r,←'' '2nd argument (optional) is caption text or "" to restore default. Omit to query current value.'
          r,←⊂'The following keywords may be used:'
          kw←'{TITLE} {WSID} {NSID} {SNSID} {PRODUCT} {VER_A} {VER_B} {VER_C} {PID} {CHARS} {BITS} {XLOC}'
          :If level∊1 3
              r,←⊂'    ',kw
              :If 3=level
                  r,←'' 'Examples:'
                  r,←⊢⊢⊂'    To change the status window''s title to "What''s going on (XX)" and include the machine size (32/64):'
                  r,←⊢⊢⊂'        ]',Cmd,' status "What''s going on ({BITS})"'
                  r,←⊢⊢⊂'    To append version number to the default session caption:'
                  r,←⊢⊢⊂'        ]',Cmd,' session {WSID} {SNSID}- {PRODUCT} {VER_A}.{VER_B}'
              :EndIf
          :Else ⍝ 2=level
              l←⊢⊂'The official window name, e.g. "Rebuild Errors"'
              l,←⊂'Workspace ID (⎕WSID), e.g. "CLEAR WS"'
              l,←⊂'Current Namespace, e.g. "#.myns"'
              l,←⊂'Current Namespace (short version), e.g. (myns)'
              l,←⊂'The name of the Dyalog product, e.g. "Dyalog APL/W - 64"'
              l,←⊂'The main version number, e.g. "14"'
              l,←⊂'The secondary version number, e.g. "1"'
              l,←⊂'The tertiary version number (currently the internal revision number), e.g. "24909"'
              l,←⊂'The process ID'
              l,←⊂'"Classic" or "Unicode"'
              l,←⊂'"32" or "64"'
              l,←⊂'The namespace currently being explored (Explorer only)'
              r,←⊢'   '∘,¨↓⍕(split kw),⍪l
          :EndIf
          :If 1≤level
              r,←⊂''
          :EndIf
      :EndIf
      :If 2≠level
          r,←⊂']???',Cmd,' ⍝ for details'
      :EndIf
      :If 3≠level
          r,←⊂']????',Cmd,' ⍝ for examples'
      :EndIf
    ∇

    ∇ r←Run(Cmd Input);key;window;str;nargs;index
      key←(1⊃4070⌶⍬),'\Captions'
      nargs←≢Input.Arguments
      r←windowValues,⍪{regGetString key ⍵}¨Windows
      :If nargs>0 ⍝ Quit if no args
          index←windowValues⍳⍬⍴Input.Arguments
          'Invalid window name'⎕SIGNAL 11/⍨index>≢windowValues
          r←2⊃index⌷r
      :AndIf nargs>1 ⍝ Quit if one arg
          (window str)←Input.Arguments
          :If 0=⍴str
              regDeleteValue key window
          :Else
              'Unbalanced {}s'⎕SIGNAL 11/⍨{(0∊¯1↑s)⍲∧/0 1∊⍨s←+\1 ¯1 0['{}'⍳⍵]}str
              regPutString key window str
          :EndIf
      :EndIf
     
    ∇

    ⍝ The following programs were modified from
    ⍝ Kai Jaeger ⋄ APL Team Ltd ⋄ http://aplteam.com
    ⍝ Homepage: http://aplwiki.com/WinReg

    ∇ regClose HANDLE;RegCloseKey
      ⎕NA'U ADVAPI32.dll.C32|RegCloseKey U'
      {}RegCloseKey HANDLE
    ∇

    HKCU←2147483649               ⍝ 'HKEY_CURRENT_USER' HEX 0x80000001
    FULL_ACCESS←983103            ⍝ HEX 0xF003F
    TrimHCU←'HKEY_CURRENT_USER\'∘{n←⍴⍺ ⋄ (n×⍺≡n↑⍵)↓⍵}

    ∇ HANDLE←regGetHandle KEY;T;Get
      T←'A*'[⎕IO+80∊⎕DR'']  ⍝ covers the case for Classic & Unicode versions
      'Get'⎕NA'I ADVAPI32.dll.C32|RegCreateKeyEx',T,' U <0T I <0T I I I >U >U'
      HANDLE←⊃2⊃Get HKCU(TrimHCU KEY)0 '' 0 FULL_ACCESS 0 0 0
    ∇

    ∇ regPutString(Path SUBKEY STRING);HANDLE;REG_SZ;set;T;uni
    ⍝ Stores the value of a Registry SUBKEY
      HANDLE←regGetHandle Path
      STRING←,STRING
      REG_SZ←1 ⍝ String data type
      T←'A*'[uni←1+80∊⎕DR'']
      'set'⎕NA'I ADVAPI32.dll.C32|RegSetValueEx',T,' U <0T I I <0T I4'
      {}set HANDLE SUBKEY 0 REG_SZ STRING(uni×1+⊃⍴STRING)
      regClose HANDLE
    ∇

    ∇ {r}←regDeleteValue(key value);handle;T;Open;Delete
     ⍝ Delete a value from the Windows Registry.
     ⍝ args are a key and a value name
     ⍝ This method normally returns either SUCCESS (0) or FILE_NOT_FOUND
     ⍝ in case the value did not exist from the start.
      T←'A*'[1+80∊⎕DR'']
      'Open'⎕NA'I ADVAPI32.dll.C32|RegOpenKeyEx',T,' U <0T I I >I'
      :If 0=⎕IO⊃,r←Open HKCU(TrimHCU key)0 FULL_ACCESS 1
          (r handle)←r
          'Delete'⎕NA'I ADVAPI32.dll.C32|RegDeleteValue',T,' U <0T[]'
          r←Delete handle value
          regClose handle
      :EndIf
    ∇

    ∇ r←regGetString(key value);handle;T;Open;rc;bufSize;Read;type
     ⍝ Read a value from the Windows Registry.
     ⍝ args are a key and a value name
      T←'A*'[1+80∊⎕DR''] ⋄ r←''
      'Open'⎕NA'I ADVAPI32.dll.C32|RegOpenKeyEx',T,' U <0T I I >I'
      :If 0=⎕IO⊃,rc←Open HKCU(TrimHCU key)0 FULL_ACCESS 1
          (r handle)←rc
          'Read'⎕NA'I ADVAPI32.dll.C32|RegQueryValueEx',T,' U <0T I =I >0T =I4'
          (rc type r bufSize)←Read handle value 0 1 1024 1024
          r/⍨←2≠rc
          regClose handle
      :EndIf
    ∇

    :ENDSECTION

    :SECTION CONSTANTS

    ⎕IO←1 ⋄ ⎕ML←1 ⋄ V15←15≤{⊢/∊⎕vfi ⍵/⍨2>+\'.'=⍵}2⊃'.'⎕wg'aplversion'

    split←{⎕ml←3 ⋄ (⍵≠' ')⊂⍵}

    Windows←'Session Editor MessageBox Explorer Rebuild_Errors Status ReFactor Event_Viewer'
    Windows,←' FindReplace ExitDialog WSSearch',V15/' SysTray'
    Windows←{b\⍵/⍨b←'_'≠⍵}¨split Windows

    windowString←'session editor messagebox explorer rebuild status refactor event findreplace exit wssearch',V15/' systray'
    windowValues←split windowString

    DESC←'Query or change window captions'

    :ENDSECTION
:EndNamespace
