﻿:namespace  EditFileAssociations ⍝ V1.21
⍝ Changes the associations of .dyapp, .dyalog and .dws files
⍝ 2015 06 30 Adam: Implemented falsely claimed feature of associating not associated types
⍝ 2015 07 01 Adam: Fixed faulty logic resulting in warped display if neither type was associated
⍝ 2015 07 01 Adam: Added missing icon, and switched wrong ones
⍝ 2015 11 12 Adam: moved help layout to framework
⍝ 2016 04 11 DanB: changed for V15
⍝ 2016 06 29 Adam: fix for 15.0 installer now not putting \ at end of install path

⍝∇:require =\WinReg.dyalog

    ⎕ml←⎕io←1
    DESC←'Changes the associations of Dyalog workspaces and applications'
    CMD←' '~⍨NAME←'File Associations'

    ∇ r←List
      →IsNotWindows/⍴r←⍬
      r←⎕NS''
      r.(Name Desc Group Parse)←CMD DESC'MSWin' ''
    ∇

    ∇ r←level Help Cmd
      r←⊂DESC,' (.dws, .dyapp and .dyalog files) using a GUI'
    ∇

    ∇ rc←Run(Cmd Args)
      RunEdit
      rc←0 0⍴''
    ∇


    :Section ControlFunctions

    ∇ RunEdit;path;curr;ivers;ipaths;f;text
      :If IsNotWindows
          ⎕←']',CMD,' applies only to Windows versions of Dyalog APL.'
          :Return
      :EndIf
     
      text←⊂'You can now specify which installed version of Dyalog APL'
      text,←⊂'should open workspace (.dws) and application (.dyapp) files.'
     
      MsgBoxInfo text
      :If ~IsUserAnAdmin
          MsgBoxAlert']',CMD,' requires Dyalog APL to be run with administrative privileges.'
          :Return
      :EndIf
     
      ivers←InstalledVersions ⍝ A vector of all installed versions of Dyalog APL from 11.0 onward
      ipaths←##.WinReg.GetString¨'HKEY_CURRENT_USER\'∘,¨ivers,¨⊂'\dyalog' ⍝ A vector of the installation directories of installed versions
      ipaths{⍺,⍵~⊃⌽⍺}¨←'\' ⍝ add trailing \ if missing
      curr←CurrentVersionUsed ivers ipaths
      curr MakeGui ivers ipaths ⍝ Generates form
     
      ⎕DQ'f'
    ∇

    ∇ (rc msg HKeys)←CurrentVersionUsed(ivers ipaths);r;path;exists;ind;mask;types
    ⍝ Determines which version of Dyalog APL, from the list of versions installed on the machine, is the
    ⍝ version already associated with files with the extesions ".dws", ".dyalog"  and ".dyapp".
      msg←3⍴⊂''
      rc←##.WinReg.DoesKeyExist¨'HKEY_CLASSES_ROOT\'∘,¨types←'dwsfile' 'dyappfile' 'dyalogfile'
      ((~rc)/msg)←⊂'Extension not defined in registry.'
     
    ⍝ dws and dyapp files are OPENed whereas dyalog files are EDITed
    ⍝ Moreover, dws files have 2 types of open: with or without ⎕LX execution
      r←##.WinReg.GetString¨HKeys←'HKEY_CLASSES_ROOT\'∘,¨types[1],¨'\shell\open'∘,¨'' ' (do not start)',¨⊂'\command\'
      r,←##.WinReg.GetString¨HKeys,←⊂'HKEY_CLASSES_ROOT\',(2⊃types),'\shell\open\command\'
      r,←##.WinReg.GetString¨HKeys,←⊂'HKEY_CLASSES_ROOT\',(3⊃types),'\shell\edit\command\'
     
      exists←Exists¨{({⍵/⍳⍴⍵}¯3⌽'.exe'⍷⍵)↑⍵}¨1↓r
      ((rc>exists)/msg)←⊂'The version of Dyalog APL associated with this file may have been moved or uninstalled.'
      path←{(⌽∨\'\'=⌽⍵)/⍵}¨1↓r
      ind←ipaths⍳path
      mask←((⍴ipaths)≥ind)∧(exists∧rc)
      msg←msg,¨mask\ivers[mask/ind]
    ∇

    ∇ Apply;ind;selpath;selver;vernum;mask
    ⍝ Identifies the versions of Dyalog installed on the machine.
    ⍝ Prompts the user to select one of these versions to associate
    ⍝ .dws, .dyalog and .dyapp files.
      mask←'f.list'⎕WG'SelItems'
      :If ~∨/mask
          →0⊣MsgBoxInfo'Please select which version of Dyalog APL'('should open ',DOPHRASE,'.')
      :EndIf
      ind←mask⍳1
      selpath←ind⊃ipaths
      selver←ind⊃ivers
      :If ~MsgBoxYN'Associate ',DOPHRASE,' with ',(AddClassic 16↓selver),'?'
          :Return
      :EndIf
      vernum←GetVersion selver
      :If applied←Set selpath vernum
          MsgBoxInfo(' were successfully associated',⍨DOPHRASE)('with ','.',⍨AddClassic 16↓selver)
      :EndIf
      ⎕EX'f'
    ∇

    ∇ rc←Set(path vernum);types;str;bin;mask;Text;dws;dyapp;dyalog;action
    ⍝ Builds data to be written to registry
    ⍝ Writes to registry.
      (dws dyapp dyalog)←DO ⋄ types←'dwsfile' 'dyappfile' 'dyalogfile'
      str←0 2⍴''
     
      :If dws
          str⍪←'.dws\' 'dwsfile'
          str⍪←'dwsfile\' 'Dyalog APL Workspace'
          str⍪←'dwsfile\shell\Open\command\'(path,'dyalog.exe "%1"')
          str⍪←'dwsfile\shell\open (do not start)\command\'(path,'dyalog.exe -x "%1"')
         ⍝ str⍪←'dwsfile\Version\'vernum
      :EndIf
     
      :If dyapp
          str⍪←'.dyapp\' 'dyappfile'
          str⍪←'dyappfile\' 'Dyalog APL Script Application'
          str⍪←'dyappfile\shell\open\command\'(path,'dyalog.exe DYAPP="%1"')
          str⍪←'dyappfile\shell\Edit\command\' 'Notepad.exe "%1"'
         ⍝ str⍪←'dyappfile\Version\'vernum
      :EndIf
     
      :If dyalog
          str⍪←'.dyalog\' 'dyalogfile'
          str⍪←'dyalogfile\' 'Dyalog APL Script Editing'
          :If 150≤⍎vernum ⋄ action←path,'dyalogrt.exe -EDITONLY "%1"'
          :Else ⋄ action←'Notepad.exe "%1"'
          :EndIf
          str⍪←'dyalogfile\shell\Open\command\'action
          str⍪←'dyalogfile\shell\Edit\command\'action
         ⍝ str⍪←'dyalogfile\Version\'vernum
      :EndIf
     
      bin←(⍪(DO/types),¨⊂'\shell\open\EditFlags'),1
     
      ⍝ Format the values
      str[;1]←(⊂'HKEY_CLASSES_ROOT\'),¨str[;1]
      bin[;1]←(⊂'HKEY_CLASSES_ROOT\'),¨bin[;1]
     
      ##.WinReg.PutString¨↓str ⍝ Sets all String Values
      ##.WinReg.PutBinary¨↓bin ⍝ Sets all Binary Values
     
      ⍝ There is also a registry key that may prevent e.g. .dyapp files from being
      ⍝ Opened (with DYALOG) if it is set:
      ⍝ HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\FileExts\.dyapp\UserChoice
     
      ⍝ Check for Completion
     
      mask←str[;2]≢¨##.WinReg.GetString¨str[;1]
      mask,←bin[;2]≢¨⊃¨##.WinReg.GetValue¨bin[;1]
     
      :If 1∊mask
          Text←'The following registry keys',⊂⎕UCS 13 10
          Text,←'were not properly written:',⊂⎕UCS 13 10
          Text,←(~mask)/(str⍪bin)[;1],¨⊂⎕UCS 13 10
          MsgBoxAlert∊Text
          rc←0
      :Else
          rc←1
      :EndIf
    ∇

    :EndSection

    :Section GUI

    ∇ curr MakeGui(ivers ipaths);Text;Y;X;vers;size;listx;neither;line1;i;type
    ⍝ Creates GUI
      vers←16↓¨ivers
     
      Text←⊂'The list below contains the versions of Dyalog APL'
      Text,←⊂'registered with Windows.'
      DO←1 1 1 ⍝ Do set workspace file association, do set application file association, do set open file association
      :If neither←⍱/2↑1⊃curr ⍝ If neither is associated
          Text,←'' 'Neither Dyalog workspace (.dws) nor application (.dyapp)'
          Text,←'files are associated with any executable.' ''
          Text,←⊂'They will both be associated with the selected version.'
     
      :ElseIf ≡/2↑2⊃curr
          Text,←'' '' 'Dyalog workspace (.dws) and application (.dyapp) files'
          Text,←('are both associated with ','.',⍨AddClassic 16↓1⊃2⊃curr)''
     
      :Else ⍝ They differ
          :For i :In ⍳2
              Text,←⊂''
              type←i⊃'workspace (.dws)' 'application (.dyapp)'
              Text,←⊂line1←'Dyalog ',type,' files are ','associated with',⍨'not '/⍨~i⊃1⊃curr
              :If i⊃1⊃curr
                  Text,←⊂'.',⍨AddClassic 16↓i⊃2⊃curr
              :Else
                  DO[i]←MsgBoxYN line1'nothing special. Would you like to associate them?'
                  Text,←⊂DO[i]/'nothing special and will remain so'
              :EndIf
          :EndFor
      :EndIf
     
      DOPHRASE←'files',⍨∊(DO[3]/'.dyalog, '),'.dyapp ' 'and ' '.dws '/⍨1⌽DO[⍳2],BOTH←∧/2↑DO ⍝ Describe upcoming changes
      Text,←⊂''
      Text,←⊂'Select the version you would like to associate',BOTH/' both'
      Text,←⊂DOPHRASE,' with, and then click apply.'
      listx←300⌊19×⍴vers
      size←(250+listx),365
      'f'⎕WC'Form'NAME('Size'size)('Coord' 'Pixel')('Sizeable' 0)('MaxButton' 0)('MinButton' 0)
      'f.fnt'⎕WC'Font'('Pname' 'Arial')('Size' 16)
      'f'⎕WS'Event'('Close' 'Cancel')
      'f'⎕WS'FontObj' 'f.fnt'
      'f.msg'⎕WC'Text'(↑Text)(15 15)
      'f.list'⎕WC'List'('Items'(AddClassic¨vers))('Posn'(200 15))('Size'(listx,335))
     
      'f.bapply'⎕WC'Button'('Caption' 'Apply')('Posn'((1⊃size)-40)100)('Size' 30 70)('Event' 'Select' 'Apply')
      'f.bcancel'⎕WC'Button'('Caption' 'Quit')('Posn'((1⊃size)-40)200)('Size' 30 70)('Event' 'Select' 'Cancel')
      ⎕NQ'f' 'Gotfocus'
    ∇

    ∇ {r}←{title}MsgBoxInfo msg;m
    ⍝ Creates a message box
      :If 0=⎕NC'title' ⋄ title←NAME ⋄ :EndIf
      'm'⎕WC'MsgBox'title msg'info'
      r←⎕DQ'm'
    ∇

    ∇ {r}←Cancel
      :If MsgBoxYN'Are you sure you want to quit?'
          ⎕EX'f' 'ipaths'
      :EndIf
    ∇

    ∇ {r}←{title}MsgBoxAlert msg;mb
      :If 0=⎕NC'title' ⋄ title←NAME ⋄ :EndIf
      'mb'⎕WC'MsgBox'('Caption'title)('Text'msg)('Style' 'Warn')('Btns' 'OK')
      r←⎕DQ'mb'
    ∇

    ∇ {r}←{title}MsgBoxYN msg;mb
      :If 0=⎕NC'title' ⋄ title←NAME ⋄ :EndIf
      'mb'⎕WC'MsgBox'('Caption'title)('Text'msg)('Style' 'Query')('Event'('MsgBtn1' 'MsgBtn2')1)
      r←'MsgBtn1'≡2⊃(⎕DQ'mb'),2⍴⊂''
    ∇
    :EndSection

    :Section Utilities

    ∇ r←Exists name
      ⍝ Does file exist?
      →0↓⍨r←name∨.≠' '
      :Trap 2
          r←⎕NEXISTS name
      :Else
          r←1
          :Trap 19 22
              ⎕NUNTIE name ⎕NTIE 0
          :Else
              r←0
          :EndTrap
      :EndTrap
    ∇
    AddClassic←  {'Unicode'≢¯7↑⍵:⍵,' Classic' ⋄ ⍵}

    ∇ r←IsUserAnAdmin;IsUserAnAdmin
     ⍝ check if this process is being "Run as Administrator"
     ⍝ BPB - nov2010
      :Trap r←0
          r←⍎⎕NA'I Shell32|IsUserAnAdmin'
      :EndTrap
    ∇

    ∇ r←InstalledVersions;sk;mask
      sk←##.WinReg.GetAllSubKeyNames'HKEY_CURRENT_USER\Software\Dyalog'
      mask←1∊¨'Dyalog APL/'∘⍷¨sk
      r←(⊂'Software\Dyalog\'),¨mask/sk
    ∇

    ∇ r←IsNotWindows
      r←'Window'≢6↑⊃'.'⎕WG'APLVersion'
    ∇

    ∇ r←IsUnicode
      r←82≠⎕DR''
    ∇

    GetVersion←{⍕10×⊃(//)⎕VFI ⍵}

    :EndSection

:Endnamespace ⍝ EditFileAssociations  $Revision: 27700 $
