﻿:Namespace outputspec ⍝ V1.02
⍝ System user Command
⍝ 2015 05 21 Adam: NS header

    ⎕io← ⎕ml←1
    outspace ← '⎕se.Dyalog.Out'    

    ∇ r←List;cmds;ver            ⍝ Name, group, short description and parsing rules
      ver←{⍎(⍵⍳'.')↑⍵}(2⊃'.'⎕WG'aplversion') ⍝ only for V14+
      →0↓⍨⍴r←cmds←(ver≥14)/'Box' 'Boxing' 'Rows' 'FindOutput'
      r←{⎕NS''}¨cmds        ⍝ space for each cmd.
      r.Name←cmds
      r.Group←3 1/'Output' 'FN'
     
      r[cmds⍳'Box' 'Boxing'].Desc←⊂'Boxing of session output'
      r[cmds⍳⊂'Rows'].Desc←'Control of row cutting and folding'
      r[cmds⍳⊂'FindOutput'].Desc←'Find where output is being produced in a program'
     
      r[cmds⍳'Box' 'Boxing'].Parse←⊂'1S -fns="off on" -style="min mid max" -trains="box tree parens" -chars="regular ascii"'
      r[cmds⍳⊂'Rows'].Parse←'1S -fns="off on" -style="long cut wrap" -dots= -fold='
      r[cmds⍳⊂'FindOutput'].Parse←'1 -includequadoutput'
    ∇

    ∇ r←Run(Cmd Input);arg;out;on;sp;cb;O;U;subs;num
      O←⍎outspace ⎕NS'Filter' 'Box' 'Rows' 'SD' 'Dft' 'pfnops' ⍝ ref.
      U←⎕SE.Dyalog.Utils                          ⍝ for lcase, disp, etc.
      (sp cb)←'SessionPrint'(outspace,'.Filter')  ⍝ callback
      r←'Was'                                     ⍝ fix for [12109]
     
      :If 0∊⎕NC↑subs←'O.B' 'O.R' 'O.F'            ⍝ spaces for Box, Rows and FindOuput.
          ⎕NS∘''¨subs                             ⍝ create state-spaces.
          O.(B R).(state fns)←⊂⊂'off'             ⍝ all states OFF
          O.B.(style trains chars)←'min' 'box' 'regular'  ⍝ default args for Boxing.
          O.R.(style fold dots)←'long' 'off' '·'  ⍝ default args for Rows.
          O.F.state←'off'                         ⍝ don't show where output is produced in programs by default
          O.F.includequadoutput←0                 ⍝ don't show ⎕ output by default
      :EndIf
     
      :Select Cmd
      :CaseList 'Box' 'Boxing'
     
          :If 0=⎕NC'O.B.trains'                   ⍝ Update Tech Preview session
              O.B.trains←'box'                    ⍝ default to boxed fns
          :EndIf
     
          arg←U.lcase⊃Input.Arguments,⊂''         ⍝ single optional argument.
     
          :Select arg
          :CaseList 'off' 'on'
              r←'Was ',U.ucase O.B.state
              O.B.state←arg
          :Case ,'?'
              r←']boxing ',U.ucase O.B.state
              r,←' -style=',(O.B.style)
              r,←' -trains=',(O.B.trains)
              r,←' -fns=',O.B.fns
              r,←' -chars=',O.B.chars
          :Case ''
              :If 0∧.≡Input.(style trains chars fns)
                  r←'Is ',U.ucase O.B.state
              :EndIf
          :Case 'reset'
              O.B.(style trains fns chars)←'min' 'box' 'off' 'regular'
              r←']boxing ',U.ucase O.B.state
              r,←' -style=',(O.B.style)
              r,←' -trains=',(O.B.trains)
              r,←' -fns=',O.B.fns
              r,←' -chars=',O.B.chars
          :Else
              r←'Arguments: ON OFF RESET ?'
          :EndSelect
     
          :If Input.style≢0                 ⍝ -style=min mid max
              r,←' -style=',O.B.style
              O.B.style←Input.style
              O.B.state←⊃(arg≡'off')⌽'on' 'off'
          :EndIf
     
          :If Input.trains≢0                ⍝ -trains=box tree parens
              r,←' -trains=',O.B.trains
              O.B.trains←Input.trains
              O.B.state←⊃(arg≡'off')⌽'on' 'off'
          :EndIf
     
          :If Input.chars≢0                 ⍝ -chars=regular ascii
              r,←' -chars=',O.B.chars
              O.B.chars←Input.chars
          :EndIf
     
          :If Input.fns≢0                   ⍝ -fns=off on
              r,←' -fns=',O.B.fns
              O.B.fns←Input.fns
          :EndIf
     
      :Case 'Rows'                          ⍝ row wrapping/cropping.
          :If 0=⎕NC'O.R.style'              ⍝ Update Tech Preview session
              :Select O.R.state
              :CaseList 'cut' 'long' 'wrap'
                  O.R.(style state←state'on')
              :Else
                  O.R.(style state←'long' 'off')
              :EndSelect
          :EndIf
          arg←U.lcase⊃Input.Arguments,⊂''   ⍝ single optional argument.
     
          :Select arg
          :CaseList 'off' 'on'
              r←'Was ',U.ucase O.R.state
              O.R.state←arg
          :Case r←''
              :If 0∧.≡Input.(style fold dots fns)
                  r←'Is ',U.ucase O.R.state
              :EndIf
          :Case ,'?'
              r←']rows ',U.ucase O.R.state
              r,←' -style=',⍕O.R.style
              r,←' -fns=',O.R.fns
              r,←' -dots=',O.R.dots
              r,←' -fold=',⍕O.R.fold
          :Else
              r←'Arguments: ON OFF ?'
          :EndSelect
     
          :If Input.style≢0                 ⍝ -style=long cut wrap
              r,←(3××⍴r)↓'was -style=',O.R.style
              O.R.style←Input.style
              O.R.state←⊃(arg≡'off')⌽'on' 'off'
          :EndIf
     
          :If Input.fold≢0                 ⍝ -fold=off, 0..9
              :If ~(⊂Input.fold)∊,¨⎕D,⊂'off'
                  r←'* Command Execution Failed: fold must be: off, 0 .. 9'
                  :Return
              :End
              r,←(3××⍴r)↓'was -fold=',O.R.fold
              O.R.fold←Input.fold
              O.R.state←⊃(arg≡'off')⌽'on' 'off'
          :EndIf
     
          :If Input.dots≢0                  ⍝ -dots=.
              r,←(3××⍴r)↓'was -dots=',O.R.dots
              O.R.dots←⊃(Input.dots~''''),'·'
          :EndIf
     
          :If Input.fns≢0                   ⍝ -fns=off on
              r,←(3××⍴r)↓'was -fns=',O.R.fns
              O.R.fns←Input.fns
          :EndIf
     
      :Case 'FindOutput'
          arg←U.lcase⊃Input.Arguments,⊂''   ⍝ single optional argument.
          :Select arg
          :CaseList 'on' 'off'
              r←'Was ',O.F.state
              O.F.state←arg ⋄ O.F.includequadoutput←Input.includequadoutput
          :Case ,'?'
              r←'Is ',O.F.state
          :Else
              r←'Arguments: ON OFF ?'
          :EndSelect
     
      :EndSelect
      on←O.(B R F).state∨.≢⊂'off'             ⍝ any active?
      '⎕se'⎕WS'Event'sp(⊃on↓0 cb)
    ∇

    ∇ r←level Help Cmd;⎕ML;lev2
      r←⍬
      :Select Cmd
      :CaseList 'Box' 'Boxing'
          r,←⊂'Boxing of default nested output'
          :If level≥1
              r,←⊂''
              r,←⊂'    ]box[ing] [arg] [-switches]  ⍝ Boxing of nested output'
              r,←⊂'     on                          ⍝ enable boxing'
              r,←⊂'     off                         ⍝ disable boxing'
              r,←⊂'     reset                       ⍝ restore factory settings'
              r,←⊂'     ?                           ⍝ query current state'
              r,←⊂'    -style=min mid max           ⍝ boxing level'
              r,←⊂'    -trains=box tree parens      ⍝ derived function display style'
              r,←⊂'    -fns=off on                  ⍝ box output from running function'
              r,←⊂'    -chars=regular ascii         ⍝ boxing chars'
              r,←⊂''
              r,←(lev2←level≥2)↓⊂'    ]???boxing                   ⍝ for more'
          :AndIf lev2
              r,←⊂''
              r,←⊂' ┌───┬──────┐  ┌→──┬──────┐  ┌→───────────────┐'
              r,←⊂' │min│boxing│  │mid│boxing│  │ ┌→──┐ ┌→─────┐ │'
              r,←⊂' └───┴──────┘  └──→┴─────→┘  │ │max│ │boxing│ │'
              r,←⊂'                             │ └───┘ └──────┘ │'
              r,←⊂'                             └∊───────────────┘'
              r,←⊂' ┌───────┬─────┐  .-----.-----.'
              r,←⊂' │regular│chars│  |ascii|chars|'
              r,←⊂' └───────┴─────┘  ''-----''-----'''
              r,←⊂''
              r,←⊂'Switch -trains controls the formatting of trains and'
              r,←⊂'derived functions'
              r,←⊂''
              r,←⊂'      ]box -trains=box'
              r,←⊂'Was -trains=box'
              r,←⊂''
              r,←⊂'      +⌿÷≢        ⍝ box-style display'
              r,←⊂'┌─────┬─┬─┐'
              r,←⊂'│┌─┬─┐│÷│≢│'
              r,←⊂'││+│⌿││ │ │'
              r,←⊂'│└─┴─┘│ │ │'
              r,←⊂'└─────┴─┴─┘'
              r,←⊂''
              r,←⊂'      ]box -t=tree'
              r,←⊂'Was -trains=box'
              r,←⊂''
              r,←⊂'      +⌿÷≢        ⍝ tree-style display'
              r,←⊂'  ┌─┼─┐'
              r,←⊂'  ⌿ ÷ ≢'
              r,←⊂'┌─┘    '
              r,←⊂'+      '
              r,←⊂'      ]box -t=parens'
              r,←⊂'Was -trains=tree'
              r,←⊂''
              r,←⊂'      +⌿÷≢        ⍝ linear, parenthesised display'
              r,←⊂'(+⌿)÷≢'
              r,←⊂''
              r,←⊂'NB: Parenthesised display will revert to the system default'
              r,←⊂'    (boxing OFF) display if any component of the derived'
              r,←⊂'    function needs more than one display line.'
              r,←⊂''
              r,←⊂'NB: With boxing on, typing the name of a function displays'
              r,←⊂'    its value (source code), rather than echoing its name.'
              r,←⊂''
              r,←⊂'By default, only the results of expressions typed into the'
              r,←⊂'session are boxed. To include output generated from _within_'
              r,←⊂'running functions, set -fns=on.'
              r,←⊂''
              r,←⊂'      ]box -fns=off'
              r,←⊂'Was -fns=off'
              r,←⊂'      {⌽⎕←⍵}''hello'' ''world'''
              r,←⊂' hello  world '
              r,←⊂'┌─────┬─────┐'
              r,←⊂'│world│hello│'
              r,←⊂'└─────┴─────┘'
              r,←⊂'      ]box -fns=on'
              r,←⊂'Was -fns=off'
              r,←⊂'      {⌽⎕←⍵}''hello'' ''world'''
              r,←⊂'┌─────┬─────┐'
              r,←⊂'│hello│world│'
              r,←⊂'└─────┴─────┘'
              r,←⊂'┌─────┬─────┐'
              r,←⊂'│world│hello│'
              r,←⊂'└─────┴─────┘'
              r,←⊂''
              r,←⊂'Argument "reset" restores switch values to their "factory'
              r,←⊂'settings": -style=min -trains=box -fns=off -chars=regular'
          :EndIf
      :Case 'Rows'
          r←⊂'Rows are cut, wrapped, scrolled or folded to fit the session window'
          :If level≥1
              r,←⊂''
              r,←⊂'    ]rows [arg] [-switches]    ⍝ Cutting or wrapping of session output'
              r,←⊂'     on                        ⍝ enable row-processing'
              r,←⊂'     off                       ⍝ restore native ⎕PW block-wrapping'
              r,←⊂'     ?                         ⍝ query current state'
              r,←⊂'    -style=long                ⍝ rows extended beyond screen width'
              r,←⊂'    -style=cut                 ⍝ rows truncated at screen width'
              r,←⊂'    -style=wrap                ⍝ each row wrapped at screen width'
              r,←⊂'    -fold=0 .. 9               ⍝ number of trailing rows after fold'
              r,←⊂'    -fold=off                  ⍝ no folding; all rows displayed'
              r,←⊂'    -fns=off on                ⍝ format output from running function'
              r,←⊂'    -dots=.                    ⍝ char to use for ...'
              r,←⊂''
              r,←(lev2←level≥2)↓⊂'    ]???rows        ⍝ for more'
          :AndIf lev2
              r,←⊂''
              r,←⊂'Row "folding" replaces rows towards the end of multi-row'
              r,←⊂'output with a line of dots so the expression that gener-'
              r,←⊂'ated the output, together with some leading and trailing'
              r,←⊂'rows, remain visible in the session window.  Row folding'
              r,←⊂'may be appropriate for session-based demonstrations.'
              r,←⊂''
              r,←⊂'      ]rows -fold=3'
              r,←⊂'Was off'
              r,←⊂''
              r,←⊂'      ⍳10 4'
              r,←⊂'┌→───┬────┬────┬────┐'
              r,←⊂'↓1 1 │1 2 │1 3 │1 4 │'
              r,←⊂'├~──→┼~──→┼~──→┼~──→┤'
              r,←⊂'│2 1 │2 2 │2 3 │2 4 │'
              r,←⊂'├~──→┼~──→┼~──→┼~──→┤'
              r,←⊂'│3 1 │3 2 │3 3 │3 4 │'
              r,←⊂'├~──→┼~──→┼~──→┼~──→┤'
              r,←⊂'·····················'
              r,←⊂'├~──→┼~──→┼~──→┼~──→┤'
              r,←⊂'│10 1│10 2│10 3│10 4│'
              r,←⊂'└~──→┴~──→┴~──→┴~──→┘'
              r,←⊂''
              r,←⊂'By default, only output from expressions typed into the session is'
              r,←⊂'processed.  To include output from running functions, set -fns=on.'
              r,←⊂''
              r,←⊂'NB: folding is incompatible with wrapping.  If both are specified,'
              r,←⊂'folding is ignored.'
          :EndIf
     
      :Case 'FindOutput'
          r←⊂'Find where programs produce output without ⎕←'
          :If level=0
              r,←⊂''
              r,←⊂'    ]findoutput arg'
              r,←⊂'     on                        ⍝ enable finding which lines produce output'
              r,←⊂'     off                       ⍝ disable process'
              r,←⊂'     ?                         ⍝ query current state'
              r,←'' 'Lines of programs producing output and NOT starting with ⎕← are displayed'
              r,←⊂'in the session preceded by a line showing which program and which line'
              r,←⊂'produced it. To include lines starting with ⎕← use the switch -includequadoutput.'
              r,←'' 'For an example type' '    ]??findoutput   '
          :Else
              r,←'Example:' '     ]findoutput  on  -includequadoutput'
              r,←⊂'    ∇ foo'
              r,←⊂'[1]   ''Line not starting with ⎕'''
              r,←⊂'[2]   ⎕←''Line with ⎕←'''
              r,←⊂'    ∇'
              r,←⊂'    foo'
              r,←⊂'>> Output from #.foo[1]'
              r,←⊂'Line not starting with ⎕'
              r,←⊂'⎕← Output from #.foo[2]'
              r,←⊂'Line with ⎕←'
     
          :EndIf
      :EndSelect
      r←↑r
    ∇

    ∇ A Filter _;susp;rand;⎕PP            ⍝ various output filters.
      ⎕PP←⍬⍴⎕RSI.⎕PP                      ⍝ current space's precision.
      :Trap 0 1000                        ⍝ catching errors & interrupts.
          susp←2=-/(,⎕STACK)⍳'*' '⎕DQ'    ⍝ execution suspended.
          susp∨←_≡1                       ⍝ simulate suspension?
          A{⍎'rand←⍺⍺' ⋄ ⍺⍺}0             ⍝ naming of fn/var "rand".
          {                               ⍝ display formatted output.
              isa←2 9∨.=⎕NC'rand'         ⍝ operand is array or namespace.
              O U←⎕SE.Dyalog.(Out Utils)  ⍝ output namespace.
              rb←'off'∘≢¨O.(R B).state    ⍝ rows and boxing states.
              oon←(2<⍴⎕XSI)∧susp<O.F.state≡'on'
              0 0 0≡rb,oon:⎕←⍵            ⍝ all off: no formatting.
              oon:⍵{0::⎕←⍺ ⋄ (fn lc)←⍵ ⋄ lc←1⌈lc
                  O.F.includequadoutput<'⎕'∊qo←'⎕←>>'/⍨2/⍲\2/'⎕←'≡2↑{(+/∧\' '=⍵)↓⍵}(1+lc)⌷⎕CR fn:⎕←⍺
                  ⎕←⍺⊣⎕←qo,' Output from ',fn,1⌽'][',⍕lc
              }3⊃¨⎕XSI ⎕LC
              fmt←{                       ⍝ formatting style.
                  box←{susp ⍺(⍵⌶Box)⍵}    ⍝ regular boxing.
                  isa:1 box ⍵             ⍝ array boxing.
                  simp←(⎕NC⊂'rand')∊3.1 3.2      ⍝ simple non-derv.
                  simp:0 box ⍵            ⍝ box-formatting array.
                  B.trains≡'tree':rand Dft 1     ⍝ derv & tree-formatting.
                  B.trains≡'box':0 box ⍵  ⍝ box-formatting array.
                  lines←⍵{∆←⍺⍺ ⋄ ≢⎕FMT ⎕CR'∆'}0  ⍝ number of lines in display
                  lines>1:⊂⍵              ⍝ multi-line train: give up.
                  ⎕THIS U.drvSrc'rand'   ⍝ parenthesised derv.
              }
              0 1≡rb:⎕←fmt ⍵              ⍝ boxing only:
              long←O.R.style≡'long'       ⍝ unrestricted screen width.
              _ ⎕PW←SD⌈32767×long         ⍝ ⎕pw temp set to screen width.
              1 0≡rb:⎕←{                  ⍝ cropping only:
                  isa:susp Rows ⎕FMT ⍵    ⍝ array: cropped.
                  ⊂⍵                      ⍝ fn: force ' ∇name'
              }⍵
              ⎕←susp Rows ⎕FMT fmt ⍵      ⍝ cropping if larger than window.
          }A                              ⍝ :: ∇ *
      :Else
          ⎕←(⎕IO+⎕EN≥1000)⊃A'...'         ⍝ error or interrupt.
      :EndTrap
    ∇

      Box←{⎕ML ⎕IO←0                      ⍝ Boxed session output.
     
          susp isa←⍺                      ⍝ from fun and array display.
          susp<B.fns≡'off':⍵              ⍝ no boxing from fn output: done.
          simp←⍵{~isa:0 ⋄ 1=≡,⍺⍺}0        ⍝ is an array and is simple.
     
          fn←{                            ⍝ formatted function.
              isa:1 box ⍵                 ⍝ top level is array.
              (⊂,⍵)∊pfns:1/⍵              ⍝ primitive fn.
              isprimop ⍵:1/⍵              ⍝ primitive op.
              cls←⍺ class ⍵               ⍝ class of function ⍵.
              cls∊+3.1 4.1:tfn ⍺          ⍝ tradfn or tradop:
              cls∊+3.2 4.2:ndfn ⍵         ⍝ named dfn or dop:
              cls∊-3.2 4.2:↑⍵             ⍝ unnamed dfn or dop:
              ~⍺ isderv ⍵:1 box ⍵         ⍝ not derived fn: give up.
              '['≡1⊃⍵:⍺{                  ⍝ +/[1 → +/[1]
                  ax←'[',(⍕2⊃⍵),']'       ⍝ formatted axis
                  (⊃⍺)ax fn(⊃⍵)ax         ⍝ axis as monadic operator
              }⍵
              ~'.'≡1⊃⍵:⍺ ∇¨⍵              ⍝ derived fn.
              0::⍺ ∇¨⍵                    ⍝ error:
              ⍙←⍎⊃⍵                       ⍝ naming left operand,
              9≠⎕NC'⍙':⍺ ∇¨⍵              ⍝ not a space ref:
              (2⊃⍺)∇ 2⊃⍵                  ⍝ dropping space-tagging.
          }                               ⍝ :: C[;] ← vr ∇ nr
     
          tfn←{'    ∇ ',(6↓¯2↓⍵),'∇'}     ⍝ nice 1970s-style display.
     
          isderv←{                        ⍝ looks like a derived fn.
              ~(⊂⍴⍵)∊,¨2 3:0              ⍝ not a 2- or 3-vector: no
              isprimop 1⊃⍵:1              ⍝ primitive operator.
              ∧/(¯2↑⍺)isfn¨¯2↑⍵:1         ⍝ fork
              4=⌊|(1⊃⍺)class 1⊃⍵          ⍝ defined operator
          }
     
          isfn←{                          ⍝ is a function?
              (⊂,⍵)∊pfns:1                ⍝ primitive fn:
              isname ⍵:1                  ⍝ for late-binding.
              3=⌊⍺ class ⍵:1              ⍝ class is 3:
              ⍺ isderv ⍵                  ⍝ derived fn.
          }                               ⍝ :: yes ← vr ∇ cr
     
          isname←{                        ⍝ valid name.
              ~ischarvec ⍵:0              ⍝ non-starter.
              0≤⎕NC ⍵                     ⍝ possible.
          }
     
          ischarvec←{(1=⍴⍴⍵)∧(⎕DR ⍵)∊82 80 160 320}  ⍝ char vector.
     
          isprimop←{                      ⍝ is a primitive operator.
              kvr←⎕UCS uni/9000+16 56 60  ⍝ key, variant, rank
              pops←'/\⌿⍀.¨∘⍨&⍣[⌶'         ⍝ classic primitive ops
              sops←'⎕S' '⎕R' '⎕OPT'       ⍝ system ops.
              (⊂⍵)∊pops,kvr,sops          ⍝ is a primitive operator.
          }
     
          ndfn←{                          ⍝ named dfn: removal of name←.
              a←⊃⍵                        ⍝ first line.
              x←a⍳'{'                     ⍝ length of 'name←'.
              ~'⍝'∊a:↑(⊂x↓a),1↓⍵          ⍝ no comment: without name.
              c←a⍳'⍝'                     ⍝ position of line[0] comment.
              d←x↓(c↑a),(x⍴' '),c↓a       ⍝ name← removed, comment adjusted.
              ↑(⊂d),1↓⍵                   ⍝ raw dfn matrix rep.
          }
     
          class←{
              ~ischarvec ⍺:¯1                     ⍝ not a char vector.
              ~'∇     ∇'≡7↑¯1⌽⍺:¯1                ⍝ not ⎕vr
              fx←(⎕NS'').{11::¯1 ⋄ ⎕NC⊂⎕FX ⍵}     ⍝ fix in tmp space.
              (cls←fx ⍵)∊3 4∘.+0.1 0.2:cls        ⍝ trad or named dfn or op:
              -fx,⊂('⍙←',⊃⍵),1↓⍵                  ⍝ unnamed dfn or failure.
          }
     
          box←{                                   ⍝ boxing of ⍵
              U B←⎕SE.Dyalog.(Utils Out.B)        ⍝ handy refs.
              style←B.style                       ⍝ min mid max
              smooth←B.chars≡'regular'            ⍝ box-drawing chars?
              simp∧~style≡'max':⍵                 ⍝ no boxing.
              style≡'min':0 smooth 0 1 ⎕PP U.disp ⍵   ⍝ min: undecorated disp
              style≡'mid':⍺ smooth 0 1 ⎕PP U.disp ⍵   ⍝ mid: ⍺-decorated disp
              ⍺:smooth ⎕PP U.display ⍵            ⍝ max: ⍺: display
              0 smooth 0 1 U.disp ⍵               ⍝ max: undecorated disp.
          }
     
          uni←80=⎕DR''                            ⍝ unicode interpreter
          pf0←'+-×÷⌊⌈|*⍟<≤=≥>≠∨∧⍱⍲!?~○'           ⍝ scalar fns.
          pf1←'⌷/⌿\⍀∊⍴↑↓⍳⊂⊃∩∪⊥⊤,⍒⍋⍉⌽⊖⌹⍕⍎⍪≡≢⍷'     ⍝ other fns.
          pfu←⎕UCS uni/8866+0 1                   ⍝ left and right tacks.
          pfns←,¨pf0,pf1,pfu                      ⍝ primitive fns.
     
          ⍙←⍺⍺ ⋄ 0 box(⊃⎕VR'⍙')fn⊃⎕NR'⍙'          ⍝ boxed output.
      }

      Dft←{⎕IO ⎕ML←0                              ⍝ Display of function tree.
     
          trav←{                                  ⍝ traverse, accumulating subtrees.
              ~(,¨2 3)∨.≡⊂⍴⍵:leaf ⍵               ⍝ not a derv or train: done.
              '['≡1⊃⍵:∇(⊃⍵)('[',(⍕2⊃⍵),']')       ⍝ axis: special treatment
              isop 1⊃⍵:{                          ⍝ derived fn:
                  2=⍴⍵:mop ⍵                      ⍝ monadic operator
                  3=⍴⍵:dop ⍵                      ⍝ dyadic operator.
              }∇¨⍵                                ⍝ formatted subtrees.
              isfn⊃⌽⍵:train ∇¨⍵                   ⍝ train:
              leaf ⍵                              ⍝ neither: give up.
          }
     
          train←{                                 ⍝ function train.
              subs←↑mesh/⍵                        ⍝ enmeshed subtrees.
              tops←apts⊃↓subs                     ⍝ anchor points.
              jpts←mid⍣(2=+/tops),tops            ⍝ joining points.
              xvec←{(3×⍵)++\⍵}jpts                ⍝ character index vector.
              deco←(¯2+⍴⍵)⊃' ── ┌┴┐ ' ' ── ┌┼┐ '  ⍝ plumbing chars.
              (xvec⊃¨⊂deco)⍪subs                  ⍝ subtree matrix.
          }
     
          mop←{                                   ⍝ operator with one operand.
              land oper←↓¨⍵                       ⍝ derived function components.
              tab←+/∧\(⊃land)∊' ┌─┐'              ⍝ indentation for top line.
              leg←dog×⌈spread÷2                   ⍝ length of dog-leg.
              pad←(tab+dog+leg)/' '               ⍝ padding for operator.
              topr←pad∘,¨oper                     ⍝ tabbed operator.
              deco←dog⊃' │  ' ' ┌─┘'              ⍝ plumbing chars.
              join←tab 1 leg dog/deco             ⍝ dog-leg or straight join.
              ↑(topr,↓join),land                  ⍝ subtree matrix.
          }
     
          dop←{                                   ⍝ operator with two operands.
              land oper rand←⍵                    ⍝ derived function components.
              subs←land mesh rand                 ⍝ merged subtrees.
              tops←mid apts⊃↓subs                 ⍝ anchor points.
              xvec←{(3×⍵)++\⍵}tops                ⍝ character index vector.
              head←xvec⊃¨⊂' ── ┌┴┐ '              ⍝ forked subtree joiner.
              otab←(head⍳'┴')/' '                 ⍝ operator padding.
              ↑(otab∘,¨↓oper),↓head⍪subs          ⍝ subtree matrix.
          }
     
          mesh←{                                  ⍝ meshed left and right subtrees.
              lbr rbr←⊂[1 2]↑(⌽⍺)⍵                ⍝ sub branches with same no of rows.
              lft rgt←{+/∧\' '=⍵}¨lbr rbr         ⍝ left and rgt adjacent blanks.
              sep←⌊/lft+rgt                       ⍝ narrowest separation.
              dpl←lft⌊sep                         ⍝ chars to drop from left subtree.
              dpr←0⌈sep-dpl                       ⍝   ..      ..      right   ..
              lvec←⌽¨dpl↓¨↓lbr                    ⍝ left rows.
              rvec←dpr↓¨(spread/' ')∘,¨↓rbr       ⍝ right rows, padded with min gap.
              trim←{(~∧\∧⌿⍵=' ')/⍵}               ⍝ trim off outer blank cols.
              ⌽trim⌽trim↑lvec,¨rvec               ⍝ meshed subtrees.
          }
     
          leaf←{                                  ⍝ format leaf.
              ⍬≡⍵:'⍬'                             ⍝ ⍬ is special
              ''≡⍵:⍉⍪''''''                       ⍝ '' is special
              1∊'←{'⍷⊃↓⍵:∇↑noname↓⍵               ⍝ dfn without name.
              dots←{(⍵⍳⍺)⊃¨⊂⍵,'·'}                ⍝ dots for blanks.
              {⍵ dots(,⍵)~' '}⎕FMT ⍵              ⍝ char matrix.
          }
     
          noname←{                                ⍝ without dfn name.
              a←⊃⍵                                ⍝ first line.
              x←a⍳'{'                             ⍝ length of 'name←'.
              ~'⍝'∊a:(⊂x↓a),1↓⍵                   ⍝ no comment: without name.
              c←a⍳'⍝'                             ⍝ position of line[0] comment.
              d←x↓(c↑a),(x⍴' '),c↓a               ⍝ name← removed, comment adjusted.
              (⊂d),1↓⍵
          }
     
          isfn←{                                  ⍝ is function?
              0=≡⍵:⍵∊pfns                         ⍝ primitive function:
              dfnop ⍵:1                           ⍝ dfn:
              ~(,¨2 3)∨.≡⊂⍴⍵:0                    ⍝ not a derv or train.
              isop 1⊃⍵:1                          ⍝ derv.
              ∇⊃⌽⍵                                ⍝ derv or train.
          }
     
          isop←{                                  ⍝ is operator?
              (⊂⍵)∊pops:1                         ⍝ primitive operator:
              ']['≡2↑¯1⌽⍵:1                       ⍝ axis: [1 2 ..]
              ~dfnop ⍵:0                          ⍝ not a d-op:
              1∊↑'⍺⍺' '⍵⍵'⍷¨⊂⍵                    ⍝ is a d-op.
          }
     
          mid←{⍵∨(⍳⍴⍵){⍺=⌊(+/⍵/⍺)÷2}⍵}            ⍝ mask with midpoint
          trim←{⍵-¯1⌽1 1⍷⍵}⍣≡                     ⍝ for ('abc',)
          apts←trim∘(~∘(∊∘' ┌─┐'))                ⍝ anchor points for sub-trees.
          dfnop←{'}'≡⊃⌽~∘' ',⍵}                   ⍝ dfn or dop
     
          pfns pops←pfnops 80=⎕DR''               ⍝ unicode-sensitive list of fns&ops.
     
          spread←1⌈⍵                              ⍝ horizontal spread.
          ⍺←1 ⋄ dog←⍺                             ⍝ dog-leg for monadic operators.
     
          fn←⍺⍺ ⋄ trav ⎕CR'fn'                    ⍝ display of function tree.
      }

      pfnops←{                                    ⍝ primmitive fns and ops.
          pf0←'+-×÷⌊⌈|*⍟<≤=≥>≠∨∧⍱⍲!?~○'           ⍝ scalar fns.
          pf1←'⌷/⌿\⍀∊⍴↑↓⍳⊂⊃∩∪⊥⊤,⍒⍋⍉⌽⊖⌹⍕⍎⍪≡≢⍷'     ⍝ other fns.
          pfu←⎕UCS ⍵/8866+0 1                     ⍝ left and right tacks.
          pfns←pf0,pf1,pfu                        ⍝ primitive fns.
     
          kvr←⎕UCS ⍵/9000+16 56 60                ⍝ key, variant, rank
          pops←'/\⌿⍀.¨∘⍨&⍣[⌶'                     ⍝ classic primitive ops
          sops←'⎕S' '⎕R' '⎕OPT'                   ⍝ system ops.
          ops←pops,kvr,sops                       ⍝ operators.
          pfns ops                                ⍝ fns & ops.
      }

      Rows←{                               ⍝ Cropped to fit session window.
          R←⎕SE.Dyalog.Out.R               ⍝ ref for params.
          ⍺<R.fns≡'off':⍵                  ⍝ no cropping from fn output: done.
          R.style≡'wrap':{0 0⍴''}{⎕←⍵}¨↓⍵  ⍝ ⎕PW-wrap each line.
          sr sc←SD⌈0,32767×R.style≡'long'  ⍝ window size
          rows cols←⍴⍵                     ⍝ raw output size.
          fold dot←R.(fold dots)           ⍝ params
          coldots←{                        ⍝ with column dots.
              cols≤sc:⍵                    ⍝ no column cropping: done.
              1≠rows:(0 ¯1↓⍵),dot          ⍝ single column of dots.
              (0 ¯3↓⍵),1 3⍴dot             ⍝ single row: stuff ...
          }
          dc←cols⌊sc                       ⍝ number of cols to display.
          nofold←(fold≡'off')∨rows≤sr-2    ⍝ no folding
          nofold:coldots rows dc↑⍵         ⍝ row cropping only.
          tail←⍎fold                       ⍝ 0..9
          t←0⌈tail⌊sr-3                    ⍝ no of trailing rows.
          h←0⌈sr-3+t                       ⍝ input + brk + fold + prompt.
          brk←sc⍴dot                       ⍝ break marker ···
          top←h sc↑⍵                       ⍝ retained upper part of output.
          bot←(t sc×¯1 1)↑⍵                ⍝    ..    lower  ..     ..
          mat←top⍪brk⍪bot                  ⍝ dots-broken matrix
          dr _←⍴mat                        ⍝ number of rows to display.
          coldots dr dc↑mat                ⍝ cropped.
      }

    ∇ rc←SD;handle;GetDlgItem;GetClientRect;dlg;r;c;v;h;_  ⍝ session window size
      :Trap 0
          ⎕NA'u user32|GetDlgItem u u'
          ⎕NA'u user32|GetClientRect u >{u u u u}'
          handle←'⎕se'⎕WG'Handle'
          dlg←GetDlgItem handle 0
          _(_ _ c r)←GetClientRect dlg(0 0 0 0)
          v h←2 ⎕NQ'⎕se' 'GetTextSize' 'X'
          rc←⌊r c÷v h
      :Else
          rc←⎕SD-1 0  ⍝ Unix version?
      :EndTrap
    ∇

:EndNamespace

 ⍝ outputspec  $Revision: 26418 $
