﻿:Namespace split ⍝ V1.03
⍝ Split a file or gather its pieces
⍝ 2015 05 22 Adam: NS header
⍝ 2015 10 27 DanB: modified split
⍝ 2015 11 02 DanB: modified collect
⍝ 2015 11 10 DanB: added compression

    ⎕IO←1 ⋄ ⎕ML←1 ⋄ AllCmds← 'Split' 'Collect' ⋄ (CR LF)←⎕UCS 13 10 ⋄ NL←⌽LF,⎕SE.SALTUtils.WIN/CR

    ∇ r←List
      r←⎕NS¨0/¨AllCmds
   ⍝ Name, group, short description and parsing rules
      r.Name←AllCmds
      r.Group←⊂'File'
      r[1].Desc←'split files into smaller files'
      r[2].Desc←'collect files into a single one'
      r.Parse←'1 -chunk= -compress -quiet -track[=]' '1 -decompress -newname= -rewrite -quiet'
    ∇

    ∇ r←Run(Cmd Input)
      :Select AllCmds⍳⊂Cmd
      :Case 1 ⍝ 'File.Split'
          r←Split Input
      :Case 2 ⍝ 'File.Collect'
          r←Collect Input
      :EndSelect
    ∇

    Compress←{2⊃2(219⌶)83⎕dr ⍵}
    Decompress←{¯2(219⌶)⍵}
    Show←{~Arg.quiet:⍵}
    If←/⍨
    ∆FI←2⊃⎕vfi
    Bid←{2 ⎕NQ'.' 'getbuildid'⍵}

    ∇ r←Collect Arg;chunk;i;out;data;f;file;old;txt;path;t;name;len;crc0;totalLen;msg;TotalLens;CRCs;n;new;verify;Expand
    ⍝ Collect files in pattern file-999
     
    ⍝ Is this an info file?
      old←new←1⊃Arg.Arguments
      path←(-⊥⍨~'/\'∊⍨old)↓old        ⍝ where files are found
      Expand←Decompress⍣Arg.decompress
      :If 0≢f←Arg.newname
          n←'='=1↑f
          f←(n/path),n↓f
          new←f,((¯1↑f)∊'/\')/(⍴path)↓old
      :EndIf
      :If verify←'.info'≡¯5↑old
         ⍝ The info file should reside with all the files to collect
         ⍝ The info file may contain a different path which was recorded
         ⍝ when the original file was split.
         ⍝ We have to account for Classic:
          txt←⎕UCS 256|83 ¯1 ⎕MAP old
          'file is empty'⎕SIGNAL 11 If 0∊⍴txt
          txt←{1↓¨(⍵=1↑⍵)⊂⍵}¯1↓LF,txt~CR  ⍝ remove CRs, cut on LFs
          n←+/≠\'"'=f←1⊃txt               ⍝ find file name without path
          len←4+⍴name←('"'=1↑t)↓t←(-⊥⍨~'/\'∊⍨t)↑t←n↑f ⍝ include -NNN
          'invalid file format (filename not found)'⎕SIGNAL 11 If 0∊⍴name
          (totalLen crc0)←{n←¯1+' '⍳⍨r←(s←⍵⍳' ')↓⍵ ⋄ ((s-1)↑⍵)(n↑s↓⍵)}t←(n+2)↓f
          old←path,name ⋄ txt←1↓txt
          Expand←Decompress⍣(Arg.decompress∨1∊'compression used'⍷t)
          new←(¯5×'.info'≡¯5↑new)↓new,(∨/'/\'=¯1↑new)/name
          'invalid sequence of files'⎕SIGNAL 11 If(⍳⍴txt)≢∆FI⍕3↑¨('-'⍳⍨1⊃txt)↓¨txt
         ⍝ Find all file lengths and CRCs
          TotalLens←∆FI⍕(n←t⍳¨' ')↑¨t←(txt⍳¨' ')↓¨txt
          CRCs←n↓¨t
      :EndIf
     
      :If Arg.rewrite
          :Trap 22
              new ⎕NERASE new ⎕NTIE 0
          :EndTrap
      :EndIf
     
      Show'Collecting files to create ',new
      out←new ⎕NCREATE 0
     
      i←0
      :Repeat
          :Trap 22
              data←0⍴f←old,'-',,'ZI3'⎕FMT i←i+1
              Show'Processing ',f,'...'
              data←83 ¯1 ⎕MAP f
              :If verify
                  msg←'file ',f,' has wrong size'
                  :If (i⊃TotalLens)≠⍴data
                      msg←'file ',f,' has wrong CRC'
                  :OrIf (i⊃CRCs)≢Bid f
                      msg ⎕SIGNAL 11
                  :EndIf
              :EndIf
              (Expand data)⎕NAPPEND out,83
          :EndTrap
      :Until 0∊⍴data
     
      t←,⎕NSIZE out ⋄ ⎕NUNTIE out
      :If verify
          ('incorrect size for file ',new)⎕SIGNAL 11 If t≢∆FI totalLen
          ('incorrect CRC for file ',new)⎕SIGNAL 11 If crc0≢Bid new
      :EndIf
      Show'Done'⊣r←(⍕i-1),' files collected.'
    ∇

    ∇ r←Split Arg;tn;chunk;i;out;data;name;ok;f;approx;fsize;csize;suffix;n;ttn;tname;add;fmt;size;drop;t
    ⍝ Split a file into several smaller files
      :Trap 22
          fsize←⎕NSIZE tn←(f←1⊃Arg.Arguments)⎕NTIE 0
      :Else
          r←'** file "',f,'" not found' ⋄ →0
      :EndTrap
     
      csize←Arg.chunk
    ⍝ Is a specific number of files wanted?
      :If csize≡approx←0
          chunk←⌈fsize÷n←10 ⍝ no, default 10 files
      :Else                 ⍝ yes, what is it?
        ⍝ If suffix T/G/M is supplied it specifies an exponent of 10
        ⍝ AND it specifies the size of the individual files
          suffix←-1<chunk←10*12 12 9 9 6 6 3 3 0['tTgGmMkK'⍳¯1↑csize]
          approx←'~'∊1↑csize ⍝ do we want them roughly equal?
          ok←(0<+/n)∧(,1)≡1⊃(n n)←⎕VFI approx↓suffix↓csize
          'Invalid number of files'⎕SIGNAL ok↓11
          chunk←⌈fsize÷n←⌈chunk×n         ⍝ number and size of files
          :If suffix≠0 ⋄ n←⌈fsize÷chunk←n ⍝ recompute size of each file
          :AndIf approx ⋄ chunk←⌈fsize÷n ⋄ :EndIf
      :EndIf
     
      :If n=.>1 999
          ⎕NUNTIE tn
          'number of files must be from 2 to 999'⎕SIGNAL 11
      :EndIf
      Show r←'Split "',f,'" into ',(⍕n),' sections of ',(⍕chunk),' bytes'
     
      ttn←0 ⋄ add←{}
      :If 0≢tname←Arg.track
          :If tname≡1 ⋄ tname←f,'.info' ⋄ :EndIf
          ttn←tname ⎕NCREATE 0 ⋄ add←{(⍵,NL)⎕NAPPEND ttn}
          t←Arg.compress/' compression used'
          add'"',f,'" ',(⍕fsize),' ',(Bid f),t,' generated on ',⍕¯1↓⎕TS
      :EndIf
     
      fmt←,('<->,ZI',⍕3⌈1+⌊10⍟n)∘⎕FMT ⍝ we allow more than 999 for later, maybe
      drop←+/∨\⌽f∊'/\'
      :For i :In ⍳n
          data←Compress⍣Arg.compress⊢⎕NREAD tn,83,chunk
          out←(name←f,fmt i)⎕NCREATE 0
          data ⎕NAPPEND out,83
          size←⎕NSIZE out
          Show name,' (',((,'CI15'⎕FMT size)~' '),' bytes)'
          ⎕NUNTIE out
          add drop↓name,' ',(⍕size),' ',Bid name
      :EndFor
      ⎕NUNTIE tn,ttn
    ∇

    ∇ r←level Help Cmd;t
      :Select AllCmds⍳⊂Cmd
      :Case 1 ⍝ 'File.Split'
          r←⍬
          r,←⊂'Split a single file into many smaller files.'
          r,←⊂' -chunk=    specifies the number of files of equal size to produce (default 10) or of maximum size.'
          r,←⊂' -compress  will compress the files using the zlib library'
          r,←⊂' -quiet     makes the command silent and no message appears while creating the files.'
          t←⊢⊢' -track[=]  tells split to track the information in file of the same name followed by .info. '
          r,←⊂t,'If -track is given a file name the information will go to that file instead.'
     
          :If level>0
              r,←'' 'Example:'
              r,←⊂'   ]split myfile -chunk=12'
              r,←⊂'will produce 12 files named myfile-001 to myfile-012'
              r,←⊂'   ]split myfile -chunk=10M'
              r,←⊂'will produce files of 10M (K, G and T also possible) each up to the last one which may be smaller than 10M.'
              r,←⊂'   ]split myfile -chunk=~10M'
              r,←⊂'will produce files all about the same size under 10M (the last one may be smaller).'
              r,←⊂'   ]split myfile -track -quiet'
              r,←⊂'will produce 10 files roughly the same size and a tracking file named myfile.info which can be used by ]collect to recreate the file.'
              r,←⊂'   ]split myfile -track -compress -chunk=1G'
              t←'will produce files of 1 Gig compressed and a tracking file named myfile.info which can be used by ]collect to recreate the file.'
              t,←' Note that you do not need a tracking file to use -compress but no verification will then be done.'
              r,←t''
          :EndIf
          r←↑r
     
      :Case 2 ⍝ 'File.Collect'
          r←⊂'Collect a series of files into a single one.'
          r,←⊂' -decompress will decompress the files as they are processed. If a tracking file is used and compression was originally specified there is no need to specify this modifier as the tracking file will record this information.'
          r,←⊂' -newname=   specifies the new name to use'
          r,←⊂' -quiet      makes the command silent and no message appears while creating the files.'
          r,←⊂' -rewrite    will erase the file if it exists'
          :If level>0
              r,←⊂'Example:'
              r,←⊂'   ]collect \tmp\mypic.jpg -rewrite -quiet'
              r,←⊂'collect all files with pattern \tmp\mypic.jpg-999'
              r,←⊂'   ]collect \tmp\pic.jpg -newname=\temp\great.jpg'
              r,←⊂'collect all files with pattern \tmp\pic.jpg-999 into \temp\great.jpg'
          :EndIf
          r←↑r
      :EndSelect
    ∇

    ∇ test track;ma;t
      ma←{p←⎕NEW ⎕SE.Parser'-chunk= -track[=] -rewrite -newname= -quiet' ⋄ p.Parse ⍵}
      {}÷'** file "dsa" not found'≡t←Split ma'dsa'
      Split ma'\tmp\from\ta.zip',track/' -t'
      Collect ma'\tmp\from\ta.zip -n=\tmp\to\'  ⍝,track/' -n='
      Collect ma'\tmp\from\ta.zip -n=\tmp\to\a.b'⍝,track/' -t'
      ⎕←'erase files now'
    ∇

:EndNamespace ⍝ split  $Revision: 25942 $