﻿:namespace wsutils ⍝ V1.02
⍝ Workspace Utilities
⍝ 2014 06 01 Unkn: DMX disabled 14/6
⍝ 2015 05 22 Adam: NS header
⍝ 2016 05 04 Modified ]defs to account for SALTed fns 

    ⎕IO←1 ⋄ ⎕ML←0   ⍝ Set those here to avoid inheriting them from outside

    AllCmds←'Defs' 'DMX' 

    ∇ r←List ⍝ this function usually returns 1 or more namespaces (here only 1)
      r←⎕NS¨0/¨AllCmds                      ⍝ create 2 commands
      r.Name←AllCmds                        ⍝ their names
      r[1].Desc←'List single line Dfns'     ⍝ descriptions
      r[2].Desc←'Display information about the last error'
      r.Group←'FN' 'Tools'
     ⍝ Parsing rules for each:
      r[1].Parse←'-string -topdown -columns'    ⍝ 'defs' accepts args to condition output
      r[2].Parse←'0 -full'
      r←1 0/r                            ⍝ DMX disabled 14/6
    ∇

    ∇ r←Run(Cmd Args);⎕USING;with;file;cr;t;full;dmx;ose;F;dir;Path;z;skipifempty;b;n;i;a;m;items;Items;names;ispat
      r←0 0⍴''
      :Select AllCmds⍳⊂Cmd
      :Case 2⊣'dmx'
          r←'⎕DMX only available starting in V13.1'
          →0⌿⍨0≡dmx←{9∊##.⎕NC'DMX':##.DMX ⋄ 0::⍵ ⋄ ⎕DMX}0
          :If dmx.EN=0 ⋄ r←'No further help available' ⋄ →0 ⋄ :EndIf
          r←1↑dmx.DM
          :If Args.full
              r,←⊂dmx.('[',Category{0∊⍴⍺:⍵ ⋄ ⍺,', ',⍵}'error #',(⍕EN),'/',(⍕ENX),']',((Vendor≢'Dyalog')/'[Vendor=',Vendor,']'),': '{⍵,⍨(0<⍴⍵)/⍺}Message)
              :If (0 0≢2↑ose←dmx.OSError)
                  t←(1+1⊃ose)⊃'C-library errno()' 'Windows GetLastError()'
                  r,←⊂'[',t,' returned ',(⍕2⊃ose),']: ',3⊃ose
              :EndIf
              r,←⊂'      Internal location: ',⍕dmx.InternalLocation
              r,←⊂'      APL Version:',(⍕'.'⎕WG'APLVersion'),' (BuildID=',(2 ⎕NQ'.' 'GetBuildID'),')'
              r,←⊂'More help at: '{⍵,⍨(0<⍴⍵)/⍺}dmx.HelpURL
              r←⍪r
     
          :Else ⍝ Not full
              m←dmx.Message
              :If (0 0≢2↑ose←dmx.OSError)
                  m,←': ',3⊃ose
              :EndIf
              :If ~0∊⍴m ⋄ r←m ⋄ :EndIf
          :EndIf
     
      :Case 1⊣'defs'
          r←n{⍺,'←',⍵}¨'',##.THIS ⎕SE.Dyalog.Utils.drvSrc¨n←##.THIS.⎕NL-3.3 4.3
          cr←↑∘⎕SE.SALTUtils.remTag¨##.THIS.⎕NR¨t←##.THIS.⎕NL-3.2 4.2 ⋄ n,←t
          t←r,t{1≠⊃⍴⍵:⍺,' (multi-line)' ⋄ ,⍵}¨cr
          Items←items←t[i←⍋↑n] ⋄ names←n[i]
          r←0 0⍴'' ⋄ b←0∊⍴a←Args.Arguments
          ispat←∨/'?*'∊⍕a
          :If Args.string∧~ispat ⋄ b∨←∨⌿1∊¨Args.Arguments∘.⍷⍕¨items
          :ElseIf ~b ⋄ b←1
              :If ispat ⋄ t←names items[⎕IO+Args.string]
                  items←items[∪↑,/t ⎕SE.SALTUtils.limRegexFind¨a]
              :Else
                  items←(items,a,¨⊂' ** UNDEFINED **')[n[i]{(⍺,⍵)⍳⍵}a]
              :EndIf
          :EndIf
          →0↓⍨(∨/b)∧⍴items
          r←⍪↑⍪{(+/∧\' '=⍵)↓⍵}¨b⌿items
          :If Args.topdown
              :If 0=≢Args.Arguments
                  r←TopDown↑Items                 ⍝ ]defs -t
              :Else
                  r←({(∧\⍵≠'←')/⍵}¨items)TopDown↑Items  ⍝ ]defs f* goo -t
              :EndIf
          :EndIf
          :If Args.columns
              r←Align r
          :EndIf
      :EndSelect
    ∇

    ∇ r←L Help Cmd
      :Select AllCmds⍳⊂Cmd
     
      :Case 1⊣'defs'
          r←,⊂'Lists definitions of 1-line dfns, dops, trains and derived functions.'
          r,←⊂'If no names are given, all definitions are displayed.'
          r,←⊂'The names may include wildcard characters:'
          r,←⊂'  * matches any sequence of characters'
          r,←⊂'  ? matches any single character.'
          r,←⊂'Modifiers:'
          r,←⊂'  -string   Show definitions that include the given string(s).'
          r,←⊂'  -topdown  Indent definitions to show the calling tree structure.'
          r,←⊂'  -columns  Display names and definitions in separate columns.'
          r,←⊂''
          r,←⊂'Examples:'
          r,←⊂'  ]defs           ⍝ show all 1-line definitions'
          r,←⊂'  ]defs df1 df2   ⍝ show only df1 and df2'
          r,←⊂'  ]defs fn*       ⍝ all definitions starting with "fn"'
          r,←⊂'  ]defs ⎕ML← -s   ⍝ all definitions containing string "⎕ML←"'
          r,←⊂'  ]defs -t        ⍝ top-down display of all definitions'
          r,←⊂'  ]defs -c        ⍝ names and code in separate columns'
          r,←⊂'  ]defs df? -t    ⍝ show df1, df2, ... and their sub-functions'
          r←↑r
     
      :Case 2⊣'dmx'
          r←'Provide information about the most recent APL error.' 'Example:' '     ÷0'
          r,←'DOMAIN ERROR' '     ÷0' '     ∧' '     ]dmx' 'Divide by zero'
          r←↑r
      :EndSelect
    ∇

      TopDown←{⎕ML←1                              ⍝ ]defs → ]defs -t
          ⍺←⊢ ⋄ all←⊃0 ⍺ 1                        ⍝ showing all defs
          vecs←{(-+/∧\⌽⍵=' ')↓⍵}¨∘↓               ⍝ vectors from matrix
          defs←vecs ⍵                             ⍝ vectors from ]defs
          names←{(∧\⍵≠'←')/⍵}¨defs                ⍝ name from each def
          patn←1↓∊{'|\b',⍵,'\b'}¨names            ⍝ whole name patterns
          hits←patn ⎕S 0 1¨defs                   ⍝ names in defs
          from←{off size←⍺ ⋄ size↑off↓⍵}          ⍝ name from offset/size
          refs←hits from¨∘⊂¨defs                  ⍝ refs per def
          tops←(⍺⊣names~↑1↓¨refs)∩names           ⍝ top-level fns
          join←{1↓↑,/'∘',⍵}                       ⍝ catenation of vectors
          clsr←{⍵∪join refs[names⍳⍵]}⍣≡           ⍝ refs transitive closure
          sort←{⍵[⍋↑⍵]}⍣all                       ⍝ sort if all defs
          roots←sort tops,all/names~clsr tops     ⍝ all root nodes
          graph←(⊂'∘',names)⍳¨(⊂roots),refs       ⍝ connected graph
          dfs←{⊃∇⍨/⌽(⊂⍺ ⍺⍺ ⍵),⍺ ⍵⍵ ⍵}             ⍝ depth-first search
          accm←{⍺,(⍺ new ⍵)/⊂⍵}                   ⍝ accumulated trees
          subs←{(⍺ new ⍵)/↑next/⍵}                ⍝ unvisited edges
          new←{~(⊃⍵)∊⊃↓⍉↑⍺}                       ⍝ node unvisited?
          next←{(⍺⊃graph),¨⍵+1}                   ⍝ depth +1 sub-trees
          indent←{(,↑⍵⍴⊂'  '),⍺⊃'∘',defs}         ⍝ ⍵-tabbed definition
          1 2↓↑↑¨indent/¨⍬ accm dfs subs 2↑⍳1     ⍝ from starting vertex.
      }

      Align←{                                    ⍝ ← and definition alignment
          vecs←↓⍵
          lens←{+/∧\~∨⌿↑'←' ' ('⍷¨⊂⍵}¨vecs       ⍝ offset of first '←' or ' ('
          arrows←↑{'←'∊⍵:' ← ' ⋄ '   '}¨vecs     ⍝ arrows separation columns
          (↑lens↑¨vecs),arrows,↑(1+lens)↓¨vecs   ⍝ name   ←  {defn
      }

:Endnamespace
 ⍝ wsutils  $Revision: 27341 $ 
